# coding: utf-8
"""

"""
import json
import typing

import flask
import flask_login
from flask_wtf import FlaskForm
from wtforms.fields import StringField, BooleanField
from wtforms.validators import DataRequired, ValidationError
from flask_babel import _

from . import frontend
from .locations import sort_location_ids_by_name, filter_location_tree_by_predicate, user_has_read_permissions_for_location, location_has_topic_ids
from .. import logic
from .utils import check_current_user_is_not_readonly
from ..logic.languages import get_language, get_language_by_lang_code
from ..logic.locations import Location, get_locations_tree
from ..logic.location_permissions import get_user_location_permissions
from ..logic.utils import get_translated_text
from ..logic.markdown_to_html import markdown_to_safe_html
from ..logic.markdown_images import mark_referenced_markdown_images_as_permanent
from ..utils import FlaskResponseT
from ..models import Permissions


class TopicsSortingForm(FlaskForm):
    encoded_order = StringField("Order-String", [DataRequired()])

    def validate_encoded_order(form, field: StringField) -> None:
        valid_topic_ids = [topic.id for topic in logic.topics.get_topics()]

        try:
            split_string = field.data.split(",")
            sorted_topic_ids = list(map(int, split_string))
            for topic_id in sorted_topic_ids:
                if topic_id not in valid_topic_ids:
                    raise ValidationError(f"{topic_id} is not a valid id")
        except ValueError:
            raise ValidationError("Invalid IDs.")

        field.data = sorted_topic_ids


@frontend.route('/topics/', methods=['GET', 'POST'])
@flask_login.login_required
def topics() -> FlaskResponseT:
    sorting_form = TopicsSortingForm()
    if sorting_form.validate_on_submit():
        check_current_user_is_not_readonly()
        logic.topics.set_topics_order(sorting_form.encoded_order.data)
    return flask.render_template(
        'topics/topics.html',
        topics=logic.topics.get_topics(),
        sorting_form=sorting_form
    )


@frontend.route('/topics/<int:topic_id>', methods=['GET', 'POST'])
@flask_login.login_required
def topic(topic_id: int) -> FlaskResponseT:
    try:
        topic = logic.topics.get_topic(
            topic_id=topic_id
        )
    except logic.errors.TopicDoesNotExistError:
        return flask.abort(404)

    if flask.request.args.get('mode') == 'edit':
        return show_topic_form(topic_id)

    if flask.current_app.config['DISABLE_INSTRUMENTS']:
        topic_instruments = []
    else:
        topic_instruments = logic.instruments.get_instruments_for_topic(topic_id=topic_id)
    topic_instruments.sort(key=lambda instrument: (
        0 if instrument.fed_id is None else 1,
        get_translated_text(instrument.name).lower()
    ))
    topic_actions = [
        action
        for action in logic.actions.get_actions_for_topic(topic_id=topic_id)
        if Permissions.READ in logic.action_permissions.get_user_action_permissions(action.id, flask_login.current_user.id) and (not action.is_hidden or flask_login.current_user.is_admin)
    ]
    topic_actions = logic.actions.sort_actions_for_user(
        actions=topic_actions,
        user_id=flask_login.current_user.id,
        sort_by_favorite=False
    )
    locations_map, locations_tree = get_locations_tree()
    location_permissions_by_id = {
        location_id: get_user_location_permissions(location_id, flask_login.current_user.id)
        for location_id in locations_map
    }

    def locations_filter(location: Location) -> bool:
        return location_has_topic_ids(location, frozenset({topic_id}))
    locations_tree = filter_location_tree_by_predicate(locations_tree, locations_map, locations_filter)
    locations_tree = filter_location_tree_by_predicate(locations_tree, locations_map, lambda location: user_has_read_permissions_for_location(location, location_permissions_by_id))
    return flask.render_template(
        'topics/topic.html',
        topic=topic,
        topic_instruments=topic_instruments,
        topic_actions=topic_actions,
        locations_map=locations_map,
        locations_tree=locations_tree,
        Permissions=Permissions,
        permissions_by_id=location_permissions_by_id,
        sort_location_ids_by_name=sort_location_ids_by_name,
        locations_filter=locations_filter
    )


@frontend.route('/topics/new', methods=['GET', 'POST'])
@flask_login.login_required
def new_topic() -> FlaskResponseT:
    if not flask_login.current_user.is_admin:
        return flask.abort(403)
    return show_topic_form(None)


class TopicForm(FlaskForm):
    translations = StringField(validators=[DataRequired()])

    show_on_frontpage = BooleanField()
    show_in_navbar = BooleanField()
    description_is_markdown = BooleanField()
    short_description_is_markdown = BooleanField()


def show_topic_form(topic_id: typing.Optional[int]) -> FlaskResponseT:
    check_current_user_is_not_readonly()
    topic_language_ids = []
    topic_form = TopicForm()

    english = logic.languages.get_language(logic.languages.Language.ENGLISH)

    def validate_string(string: str) -> bool:
        try:
            return 0 < len(string) < 100
        except Exception:
            return False

    def validate_strings(strings: typing.Sequence[str]) -> bool:
        try:
            result = [validate_string(string) for string in strings]
            for temp in result:
                if temp is False:
                    return False
        except Exception:
            return False

        return True

    class TopicTranslation(typing.TypedDict):
        language_id: typing.Union[int, str]
        lang_name: str
        name: str
        description: str
        short_description: str

    translations: typing.List[TopicTranslation] = []

    if topic_id is not None:
        try:
            topic = logic.topics.get_topic(topic_id)
        except logic.errors.TopicDoesNotExistError:
            return flask.abort(404)
        for lang_code, name in topic.name.items():
            lang_id = get_language_by_lang_code(lang_code).id
            translation: TopicTranslation = {
                'language_id': str(lang_id),
                'lang_name': get_translated_text(get_language(lang_id).names),
                'name': name,
                'description': '',
                'short_description': ''
            }
            translations.append(translation)

        for lang_code, description in topic.description.items():
            if lang_code == '':
                continue
            lang_id = get_language_by_lang_code(lang_code).id
            for translation in translations:
                if str(lang_id) == translation['language_id']:
                    translation['description'] = description
                    break
            else:
                translation = {
                    'language_id': str(lang_id),
                    'lang_name': get_translated_text(get_language(lang_id).names),
                    'name': '',
                    'description': description,
                    'short_description': ''
                }
                translations.append(translation)

        for lang_code, short_description in topic.short_description.items():
            if lang_code == '':
                continue
            lang_id = get_language_by_lang_code(lang_code).id
            for translation in translations:
                if str(lang_id) == translation['language_id']:
                    translation['short_description'] = short_description
                    break
            else:
                translation = {
                    'language_id': str(lang_id),
                    'lang_name': get_translated_text(get_language(lang_id).names),
                    'name': '',
                    'description': '',
                    'short_description': short_description
                }
                translations.append(translation)
        if 'action_submit' not in flask.request.form:
            topic_language_ids = [int(translation['language_id']) for translation in translations]
            topic_form.show_on_frontpage.data = topic.show_on_frontpage
            topic_form.show_in_navbar.data = topic.show_in_navbar
            topic_form.description_is_markdown.data = topic.description_is_markdown
            topic_form.short_description_is_markdown.data = topic.short_description_is_markdown

    if topic_form.validate_on_submit():
        try:
            translation_data = json.loads(topic_form.translations.data)
        except Exception:
            flask.flash(_('There was a problem saving the topic data.'), 'error')
            return flask.render_template(
                'topics/topic_form.html',
                topic_form=topic_form,
                topic_translations=translations,
                languages=logic.languages.get_languages(),
                ENGLISH=english,
                submit_text=_('Create') if topic_id is None else _('Save')
            )
        else:
            valid_translation_keys = {'language_id', 'name', 'description', 'short_description'}
            if not isinstance(translation_data, list):
                translation_data = ()
            for translation in translation_data:
                if not isinstance(translation, dict):
                    flask.flash(_('Please fill out the form.'), 'error')
                    return flask.render_template(
                        'topics/topic_form.html',
                        topic_form=topic_form,
                        topic_translations=translations,
                        languages=logic.languages.get_languages(),
                        ENGLISH=english,
                        submit_text=_('Create') if topic_id is None else _('Save')
                    )

                translation_keys = set(translation.keys())
                translation_keys -= {'lang_name'}
                if translation_keys != valid_translation_keys:
                    flask.flash(_('Please fill out the form.'), 'error')
                    return flask.render_template(
                        'topics/topic_form.html',
                        topic_form=topic_form,
                        topic_translations=translations,
                        languages=logic.languages.get_languages(),
                        ENGLISH=english,
                        submit_text=_('Create') if topic_id is None else _('Save')
                    )

                name = translation['name'].strip()

                if not validate_strings([name]):
                    flask.flash(_('The name is required.'), 'error')
                    return flask.render_template(
                        'topics/topic_form.html',
                        topic_form=topic_form,
                        topic_translations=translations,
                        languages=logic.languages.get_languages(),
                        ENGLISH=english,
                        submit_text=_('Create') if topic_id is None else _('Save')
                    )

            names: typing.Dict[str, str] = {}
            descriptions: typing.Dict[str, str] = {}
            short_descriptions: typing.Dict[str, str] = {}

            for translation in translation_data:
                language_id = int(translation['language_id'])
                language_code = get_language(language_id).lang_code
                names[language_code] = translation['name'].strip()
                descriptions[language_code] = translation['description'].strip()
                short_descriptions[language_code] = translation['short_description'].strip()

                if topic_form.description_is_markdown.data:
                    description_as_html = markdown_to_safe_html(descriptions[language_code], anchor_prefix="topic-description")
                    mark_referenced_markdown_images_as_permanent(description_as_html)
                if topic_form.short_description_is_markdown.data:
                    short_description_as_html = markdown_to_safe_html(short_descriptions[language_code], anchor_prefix="topic-short-description")
                    mark_referenced_markdown_images_as_permanent(short_description_as_html)

            if topic_id is None:
                topic = logic.topics.create_topic(
                    name=names,
                    description=descriptions,
                    short_description=short_descriptions,
                    show_on_frontpage=topic_form.show_on_frontpage.data,
                    show_in_navbar=topic_form.show_in_navbar.data,
                    description_is_markdown=topic_form.description_is_markdown.data,
                    short_description_is_markdown=topic_form.short_description_is_markdown.data
                )
                logic.topics.add_topic_to_order(topic)
            else:
                topic = logic.topics.update_topic(
                    topic_id=topic_id,
                    name=names,
                    description=descriptions,
                    short_description=short_descriptions,
                    show_on_frontpage=topic_form.show_on_frontpage.data,
                    show_in_navbar=topic_form.show_in_navbar.data,
                    description_is_markdown=topic_form.description_is_markdown.data,
                    short_description_is_markdown=topic_form.short_description_is_markdown.data
                )

        return flask.redirect(flask.url_for('.topic', topic_id=topic.id))

    return flask.render_template(
        'topics/topic_form.html',
        topic_form=topic_form,
        topic_translations=translations,
        topic_language_ids=topic_language_ids,
        languages=logic.languages.get_languages(),
        ENGLISH=english,
        submit_text=_('Create') if topic_id is None else _('Save')
    )



import base64
import typing

import flask

from .components import _get_or_create_component_id
from .utils import _get_id, _get_uuid, _get_bool, _get_str, _get_dict
from ..languages import get_languages, get_language, get_language_by_lang_code
from ..instruments import get_instrument, get_mutable_instrument, create_instrument, Instrument
from ..instrument_translations import set_instrument_translation, get_instrument_translations_for_instrument
from ..markdown_images import get_markdown_image, find_referenced_markdown_images
from ..components import Component
from .. import errors, fed_logs, markdown_to_html
from ... import db


class InstrumentRef(typing.TypedDict):
    instrument_id: int
    component_uuid: str


class InstrumentTranslationData(typing.TypedDict):
    language_id: int
    name: str
    description: str
    short_description: str
    notes: str


class InstrumentData(typing.TypedDict):
    fed_id: int
    component_uuid: str
    description_is_markdown: bool
    short_description_is_markdown: bool
    notes_is_markdown: bool
    is_hidden: bool
    users_can_create_log_entries: bool
    users_can_view_log_entries: bool
    create_log_entry_default: bool
    translations: typing.List[InstrumentTranslationData]


class SharedInstrumentTranslationData(typing.TypedDict):
    name: str
    description: str
    short_description: str
    notes: str


class SharedInstrumentData(typing.TypedDict):
    instrument_id: int
    component_uuid: str
    description_is_markdown: bool
    short_description_is_markdown: bool
    notes_is_markdown: bool
    is_hidden: bool
    translations: typing.Optional[typing.Dict[str, SharedInstrumentTranslationData]]


def parse_instrument(
        instrument_data: typing.Dict[str, typing.Any]
) -> InstrumentData:
    fed_id = _get_id(instrument_data.get('instrument_id'))
    uuid = _get_uuid(instrument_data.get('component_uuid'))
    if uuid == flask.current_app.config['FEDERATION_UUID']:
        # do not accept updates for own data
        raise errors.InvalidDataExportError(f'Invalid update for local instrument {fed_id}')
    result = InstrumentData(
        fed_id=fed_id,
        component_uuid=uuid,
        description_is_markdown=_get_bool(instrument_data.get('description_is_markdown'), default=False),
        is_hidden=_get_bool(instrument_data.get('is_hidden'), default=False),
        short_description_is_markdown=_get_bool(instrument_data.get('short_description_is_markdown'), default=False),
        notes_is_markdown=_get_bool(instrument_data.get('notes_is_markdown'), default=False),
        users_can_create_log_entries=False,
        users_can_view_log_entries=False,
        create_log_entry_default=False,
        translations=[]
    )

    allowed_language_ids = [language.id for language in get_languages(only_enabled_for_input=False)]

    translation_data = _get_dict(instrument_data.get('translations'))
    if translation_data is not None:
        for lang_code, translation in translation_data.items():
            try:
                language = get_language_by_lang_code(lang_code.lower())
            except errors.LanguageDoesNotExistError:
                continue
            language_id = language.id
            if language_id not in allowed_language_ids:
                continue

            result['translations'].append(InstrumentTranslationData(
                language_id=language_id,
                name=_get_str(translation.get('name'), default=''),
                description=_get_str(translation.get('description'), default=''),
                short_description=_get_str(translation.get('short_description'), default=''),
                notes=_get_str(translation.get('notes'), default='')
            ))
    return result


def import_instrument(
        instrument_data: InstrumentData,
        component: Component
) -> Instrument:
    component_id = _get_or_create_component_id(instrument_data['component_uuid'])
    # component_id will only be None if this would import a local instrument
    assert component_id is not None
    try:
        mutable_instrument = get_mutable_instrument(instrument_data['fed_id'], component_id)

        ignored_keys = {
            'fed_id',
            'component_uuid',
            'translations'
        }
        if any(
                value != getattr(mutable_instrument, key)
                for key, value in instrument_data.items()
                if key not in ignored_keys
        ):
            mutable_instrument.description_is_markdown = instrument_data['description_is_markdown']
            mutable_instrument.is_hidden = instrument_data['is_hidden']
            mutable_instrument.short_description_is_markdown = instrument_data['short_description_is_markdown']
            db.session.commit()
            fed_logs.update_instrument(mutable_instrument.id, component.id)
        instrument = Instrument.from_database(mutable_instrument)
    except errors.InstrumentDoesNotExistError:
        instrument = create_instrument(
            fed_id=instrument_data['fed_id'],
            component_id=component_id,
            description_is_markdown=instrument_data['description_is_markdown'],
            is_hidden=instrument_data['is_hidden'],
            short_description_is_markdown=instrument_data['short_description_is_markdown'],
            notes_is_markdown=instrument_data['notes_is_markdown'],
            users_can_create_log_entries=instrument_data['users_can_create_log_entries'],
            users_can_view_log_entries=instrument_data['users_can_view_log_entries'],
            create_log_entry_default=instrument_data['create_log_entry_default']
        )
        fed_logs.import_instrument(instrument.id, component.id)

    for instrument_translation in instrument_data['translations']:
        set_instrument_translation(
            instrument_id=instrument.id,
            language_id=instrument_translation['language_id'],
            name=instrument_translation['name'],
            description=instrument_translation['description'],
            short_description=instrument_translation['short_description'],
            notes=instrument_translation['notes']
        )

    return instrument


def parse_import_instrument(
        instrument_data: typing.Dict[str, typing.Any],
        component: Component
) -> Instrument:
    return import_instrument(parse_instrument(instrument_data), component)


def _get_or_create_instrument_id(
        instrument_data: typing.Optional[InstrumentRef]
) -> typing.Optional[int]:
    if instrument_data is None:
        return None
    component_id = _get_or_create_component_id(instrument_data['component_uuid'])
    try:
        instrument = get_instrument(instrument_data['instrument_id'], component_id)
    except errors.InstrumentDoesNotExistError:
        assert component_id is not None
        instrument = create_instrument(fed_id=instrument_data['instrument_id'], component_id=component_id)
        fed_logs.create_ref_instrument(instrument.id, component_id)
    return instrument.id


def _parse_instrument_ref(
        instrument_data: typing.Optional[typing.Union[InstrumentRef, typing.Dict[str, typing.Any]]]
) -> typing.Optional[InstrumentRef]:
    if instrument_data is None:
        return None
    instrument_id = _get_id(instrument_data.get('instrument_id'))
    component_uuid = _get_uuid(instrument_data.get('component_uuid'))
    return InstrumentRef(
        instrument_id=instrument_id,
        component_uuid=component_uuid
    )


def shared_instrument_preprocessor(
        instrument_id: int,
        _component: Component,
        _refs: typing.List[typing.Tuple[str, int]],
        markdown_images: typing.Dict[str, str]
) -> typing.Optional[SharedInstrumentData]:
    instrument = get_instrument(instrument_id)
    if instrument.component_id is not None:
        return None
    translations_data = {}
    try:
        translations = get_instrument_translations_for_instrument(instrument_id)
        for translation in translations:
            lang_code = get_language(translation.language_id).lang_code

            description = translation.description
            markdown_as_html = markdown_to_html.markdown_to_safe_html(description)
            for file_name in find_referenced_markdown_images(markdown_as_html):
                markdown_image_b = get_markdown_image(file_name, None)
                description = description.replace('/markdown_images/' + file_name, '/markdown_images/' + flask.current_app.config['FEDERATION_UUID'] + '/' + file_name)
                if markdown_image_b is not None and file_name not in markdown_images:
                    markdown_images[file_name] = base64.b64encode(markdown_image_b).decode('utf-8')

            short_description = translation.short_description
            markdown_as_html = markdown_to_html.markdown_to_safe_html(short_description)
            for file_name in find_referenced_markdown_images(markdown_as_html):
                markdown_image_b = get_markdown_image(file_name, None)
                short_description = short_description.replace('/markdown_images/' + file_name, '/markdown_images/' + flask.current_app.config['FEDERATION_UUID'] + '/' + file_name)
                if markdown_image_b is not None and file_name not in markdown_images:
                    markdown_images[file_name] = base64.b64encode(markdown_image_b).decode('utf-8')

            notes = translation.notes
            markdown_as_html = markdown_to_html.markdown_to_safe_html(notes)
            for file_name in find_referenced_markdown_images(markdown_as_html):
                markdown_image_b = get_markdown_image(file_name, None)
                notes = notes.replace('/markdown_images/' + file_name, '/markdown_images/' + flask.current_app.config['FEDERATION_UUID'] + '/' + file_name)
                if markdown_image_b is not None and file_name not in markdown_images:
                    markdown_images[file_name] = base64.b64encode(markdown_image_b).decode('utf-8')

            translations_data[lang_code] = SharedInstrumentTranslationData(
                name=translation.name,
                description=description,
                short_description=short_description,
                notes=notes
            )
    except errors.InstrumentTranslationDoesNotExistError:
        translations_data = {}
    return SharedInstrumentData(
        instrument_id=instrument.id if instrument.fed_id is None else instrument.fed_id,
        component_uuid=flask.current_app.config['FEDERATION_UUID'] if instrument.component is None else instrument.component.uuid,
        description_is_markdown=instrument.description_is_markdown,
        notes_is_markdown=instrument.notes_is_markdown,
        is_hidden=instrument.is_hidden,
        short_description_is_markdown=instrument.short_description_is_markdown,
        translations=translations_data if translations_data else None
    )



# coding: utf-8
"""

"""
import typing
import datetime

import flask
from sqlalchemy.orm import Mapped, Query

from .. import db
from .utils import Model


class DownloadServiceJobFile(Model):
    __tablename__ = 'download_service_job_files'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True, autoincrement=True)
    object_id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    file_id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    creation: Mapped[datetime.datetime] = db.Column(db.TIMESTAMP(timezone=True), nullable=False)
    expiration: Mapped[datetime.datetime] = db.Column(db.TIMESTAMP(timezone=True), nullable=False)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["DownloadServiceJobFile"]]

    __table_args__ = (
        db.ForeignKeyConstraint(['object_id', 'file_id'], ['files.object_id', 'files.id']),
    )

    def __init__(
            self,
            job_id: typing.Optional[int],
            object_id: int,
            file_id: int
    ) -> None:
        super().__init__(
            id=job_id,
            object_id=object_id,
            file_id=file_id,
            creation=datetime.datetime.now(datetime.timezone.utc),
            expiration=datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(seconds=flask.current_app.config['DOWNLOAD_SERVICE_TIME_LIMIT'])
        )

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id}, object_id={self.object_id}, file_id={self.file_id}'



import base64
import typing

import flask

from .components import _get_or_create_component_id
from .utils import _get_id, _get_uuid, _get_bool, _get_str, _get_dict
from .action_types import _parse_action_type_ref, _get_or_create_action_type_id, ActionTypeRef
from .instruments import _parse_instrument_ref, _get_or_create_instrument_id, InstrumentRef
from .users import _parse_user_ref, _get_or_create_user_id, UserRef
from ..action_permissions import set_action_permissions_for_all_users
from ..actions import get_action, get_mutable_action, create_action, Action
from ..action_translations import set_action_translation, get_action_translations_for_action
from ..languages import get_languages, get_language, get_language_by_lang_code, get_language_codes
from ..instruments import get_instrument
from ..markdown_images import get_markdown_image, find_referenced_markdown_images
from ..components import Component, get_component, get_component_by_uuid
from ..users import get_user
from ..schemas.validate_schema import validate_schema
from ..schemas.utils import schema_iter
from .. import errors, fed_logs, markdown_to_html
from ... import db
from ...models import Permissions


class ActionRef(typing.TypedDict):
    action_id: int
    component_uuid: str


class ActionTranslationData(typing.TypedDict):
    language_id: int
    name: str
    description: str
    short_description: str


class ActionData(typing.TypedDict):
    fed_id: int
    component_uuid: str
    action_type: ActionTypeRef
    schema: typing.Optional[typing.Dict[str, typing.Any]]
    instrument: typing.Optional[InstrumentRef]
    user: typing.Optional[UserRef]
    description_is_markdown: bool
    short_description_is_markdown: bool
    is_hidden: bool
    translations: typing.List[ActionTranslationData]
    admin_only: bool
    disable_create_objects: bool
    import_notes: typing.List[str]


class SharedActionTranslationData(typing.TypedDict):
    name: str
    description: str
    short_description: str


class SharedActionData(typing.TypedDict):
    action_id: int
    component_uuid: str
    action_type: typing.Optional[ActionTypeRef]
    schema: typing.Optional[typing.Dict[str, typing.Any]]
    instrument: typing.Optional[InstrumentRef]
    user: typing.Optional[UserRef]
    description_is_markdown: bool
    short_description_is_markdown: bool
    is_hidden: bool
    translations: typing.Optional[typing.Dict[str, SharedActionTranslationData]]
    admin_only: bool
    disable_create_objects: bool


def parse_action(
        action_data: typing.Dict[str, typing.Any]
) -> ActionData:
    import_notes = []
    fed_id = _get_id(action_data.get('action_id'))
    uuid = _get_uuid(action_data.get('component_uuid'))
    if uuid == flask.current_app.config['FEDERATION_UUID']:
        # do not accept updates for own data
        raise errors.InvalidDataExportError(f'Invalid update for local action {fed_id}')
    schema: typing.Optional[typing.Dict[str, typing.Any]] = _get_dict(action_data.get('schema'))
    if schema is not None:
        _parse_schema(schema)
        try:
            validate_schema(schema, strict=False)
        except errors.ValidationError as e:
            schema = None
            import_notes.append(f'Invalid schema in action #{fed_id} @ {uuid} ({e})')

    result = ActionData(
        fed_id=fed_id,
        component_uuid=uuid,
        action_type=_parse_action_type_ref(_get_dict(action_data.get('action_type'), mandatory=True)),
        schema=schema,
        instrument=_parse_instrument_ref(_get_dict(action_data.get('instrument'))),
        user=_parse_user_ref(_get_dict(action_data.get('user'))),
        description_is_markdown=_get_bool(action_data.get('description_is_markdown'), default=False),
        is_hidden=_get_bool(action_data.get('is_hidden'), default=False),
        short_description_is_markdown=_get_bool(action_data.get('short_description_is_markdown'), default=False),
        translations=[],
        admin_only=_get_bool(action_data.get('admin_only'), default=True),
        disable_create_objects=_get_bool(action_data.get('disable_create_objects'), default=True),
        import_notes=import_notes
    )

    allowed_language_ids = [language.id for language in get_languages(only_enabled_for_input=False)]

    translation_data = _get_dict(action_data.get('translations'))
    if translation_data is not None:
        for lang_code, translation in translation_data.items():
            try:
                language = get_language_by_lang_code(lang_code.lower())
            except errors.LanguageDoesNotExistError:
                continue
            language_id = language.id
            if language_id not in allowed_language_ids:
                continue

            result['translations'].append(ActionTranslationData(
                language_id=language_id,
                name=_get_str(translation.get('name'), default=''),
                description=_get_str(translation.get('description'), default=''),
                short_description=_get_str(translation.get('short_description'), default='')
            ))
    return result


def import_action(
        action_data: ActionData,
        component: Component
) -> Action:
    component_id = _get_or_create_component_id(action_data['component_uuid'])
    # component_id will only be None if this would import a local action
    assert component_id is not None

    action_type_id = _get_or_create_action_type_id(action_data['action_type'])
    instrument_id = _get_or_create_instrument_id(action_data['instrument'])
    user_id = _get_or_create_user_id(action_data['user'])

    schema = action_data['schema']
    if schema is not None:
        for _, property_schema in schema_iter(schema, filter_property_types={'object'}):
            if 'template' in property_schema:
                try:
                    c = get_component_by_uuid(property_schema['template']['component_uuid'])
                    property_schema['template'] = get_action(property_schema['template']['action_id'], c.id).id
                except errors.ActionDoesNotExistError:
                    pass
                except errors.ComponentDoesNotExistError:
                    pass

    try:
        mutable_action = get_mutable_action(action_data['fed_id'], component_id)
        ignored_keys = {
            'fed_id',
            'component_uuid',
            'action_type',
            'instrument',
            'user',
            'schema',
            'translations',
            'import_notes',
            'disable_create_objects'
        }
        if any(
                value != getattr(mutable_action, key)
                for key, value in action_data.items()
                if key not in ignored_keys
        ) or mutable_action.type_id != action_type_id or mutable_action.instrument_id != instrument_id or mutable_action.user_id != user_id or mutable_action.schema != schema:
            mutable_action.type_id = action_type_id
            mutable_action.schema = schema
            mutable_action.instrument_id = instrument_id
            mutable_action.user_id = user_id
            mutable_action.description_is_markdown = action_data['description_is_markdown']
            mutable_action.is_hidden = action_data['is_hidden']
            mutable_action.short_description_is_markdown = action_data['short_description_is_markdown']
            mutable_action.admin_only = action_data['admin_only']
            db.session.commit()
            fed_logs.update_action(mutable_action.id, component.id, action_data.get('import_notes', []))
        action = Action.from_database(mutable_action)
    except errors.ActionDoesNotExistError:
        action = create_action(
            fed_id=action_data['fed_id'],
            component_id=component_id,
            action_type_id=action_type_id,
            schema=schema,
            instrument_id=instrument_id,
            user_id=user_id,
            description_is_markdown=action_data['description_is_markdown'],
            is_hidden=action_data['is_hidden'],
            short_description_is_markdown=action_data['short_description_is_markdown'],
            admin_only=action_data.get('admin_only', True),
            disable_create_objects=True,
            strict_schema_validation=False,
        )
        fed_logs.import_action(action.id, component.id, action_data.get('import_notes', []))
        set_action_permissions_for_all_users(action.id, Permissions.READ)

    for action_translation in action_data['translations']:
        set_action_translation(
            action_id=action.id,
            language_id=action_translation['language_id'],
            name=action_translation['name'],
            description=action_translation['description'],
            short_description=action_translation['short_description']
        )
    return action


def parse_import_action(
        action_data: typing.Dict[str, typing.Any],
        component: Component
) -> Action:
    return import_action(parse_action(action_data), component)


def _parse_action_ref(
        action_data: typing.Optional[typing.Union[ActionRef, typing.Dict[str, typing.Any]]]
) -> typing.Optional[ActionRef]:
    if action_data is None:
        return None
    action_id = _get_id(action_data.get('action_id'))
    component_uuid = _get_uuid(action_data.get('component_uuid'))
    return ActionRef(
        action_id=action_id,
        component_uuid=component_uuid
    )


def _get_or_create_action_id(
        action_data: typing.Optional[ActionRef]
) -> typing.Optional[int]:
    if action_data is None:
        return None
    component_id = _get_or_create_component_id(action_data['component_uuid'])
    try:
        action = get_action(action_data['action_id'], component_id)
    except errors.ActionDoesNotExistError:
        assert component_id is not None
        action = create_action(action_type_id=None, schema=None, fed_id=action_data['action_id'], component_id=component_id)
        fed_logs.create_ref_action(action.id, component_id)
    return action.id


def shared_action_preprocessor(
        action_id: int,
        _component: Component,
        refs: typing.List[typing.Tuple[str, int]],
        markdown_images: typing.Dict[str, str]
) -> typing.Optional[SharedActionData]:
    action = get_action(action_id)
    if action.component_id is not None:
        return None
    translations_data = {}

    try:
        translations = get_action_translations_for_action(action_id)
        for translation in translations:
            description = translation.description
            markdown_as_html = markdown_to_html.markdown_to_safe_html(description)
            for file_name in find_referenced_markdown_images(markdown_as_html):
                markdown_image_b = get_markdown_image(file_name, None)
                description = description.replace('/markdown_images/' + file_name, '/markdown_images/' + flask.current_app.config['FEDERATION_UUID'] + '/' + file_name)
                if markdown_image_b is not None and file_name not in markdown_images:
                    markdown_images[file_name] = base64.b64encode(markdown_image_b).decode('utf-8')

            short_description = translation.short_description
            markdown_as_html = markdown_to_html.markdown_to_safe_html(short_description)
            for file_name in find_referenced_markdown_images(markdown_as_html):
                markdown_image_b = get_markdown_image(file_name, None)
                short_description = short_description.replace('/markdown_images/' + file_name, '/markdown_images/' + flask.current_app.config['FEDERATION_UUID'] + '/' + file_name)
                if markdown_image_b is not None and file_name not in markdown_images:
                    markdown_images[file_name] = base64.b64encode(markdown_image_b).decode('utf-8')

            lang_code = get_language(translation.language_id).lang_code
            translations_data[lang_code] = SharedActionTranslationData(
                name=translation.name,
                description=description,
                short_description=short_description
            )
    except errors.ActionTranslationDoesNotExistError:
        translations_data = {}

    if action.instrument_id is not None and ('instruments', action.instrument_id) not in refs:
        refs.append(('instruments', action.instrument_id))
    if action.user_id is not None and ('users', action.user_id) not in refs:
        refs.append(('users', action.user_id))
    if action.instrument_id is None or flask.current_app.config['DISABLE_INSTRUMENTS']:
        instrument = None
    else:
        i = get_instrument(action.instrument_id)
        if i.component is None or i.fed_id is None:
            instrument = InstrumentRef(
                instrument_id=i.id,
                component_uuid=flask.current_app.config['FEDERATION_UUID']
            )
        else:
            instrument = InstrumentRef(
                instrument_id=i.fed_id,
                component_uuid=i.component.uuid
            )
    if action.type is None:
        action_type = None
    else:
        if action.type.component_id is None or action.type.fed_id is None:
            action_type = ActionTypeRef(
                action_type_id=action.type.id,
                component_uuid=flask.current_app.config['FEDERATION_UUID']
            )
        else:
            comp = get_component(action.type.component_id)
            action_type = ActionTypeRef(
                action_type_id=action.type.fed_id,
                component_uuid=comp.uuid
            )
        if ('action_types', action.type.id) not in refs:
            refs.append(('action_types', action.type.id))
    if action.user_id is None:
        user = None
    else:
        u = get_user(action.user_id)
        if u.component is None or u.fed_id is None:
            user = UserRef(
                user_id=u.id,
                component_uuid=flask.current_app.config['FEDERATION_UUID']
            )
        else:
            user = UserRef(
                user_id=u.fed_id,
                component_uuid=u.component.uuid
            )
    if action.schema is None:
        schema = None
    else:
        schema = action.schema.copy()
        schema_entry_preprocessor(schema, refs)
    return SharedActionData(
        action_id=action.id if action.fed_id is None else action.fed_id,
        component_uuid=flask.current_app.config['FEDERATION_UUID'] if action.component is None else action.component.uuid,
        action_type=action_type,
        instrument=instrument,
        schema=schema,
        user=user,
        description_is_markdown=action.description_is_markdown,
        is_hidden=action.is_hidden,
        short_description_is_markdown=action.short_description_is_markdown,
        translations=translations_data if translations_data else None,
        admin_only=action.admin_only,
        disable_create_objects=action.disable_create_objects,
    )


def _parse_schema(
        schema: typing.Dict[str, typing.Any],
        path: typing.Optional[typing.List[str]] = None
) -> None:
    if schema is None:
        return
    if path is None:
        path = []
    all_language_codes = get_language_codes()
    for key in ['title', 'placeholder', 'default', 'note']:
        if key in schema and isinstance(schema[key], dict):
            for lang_code in list(schema[key].keys()):
                if lang_code not in all_language_codes:
                    del schema[key][lang_code]

    if 'languages' in schema and schema['languages'] != 'all':
        if not isinstance(schema['languages'], list):
            raise errors.InvalidDataExportError('Invalid schema')
        for language in schema['languages']:
            if language not in all_language_codes:
                schema['languages'].remove(language)

    if 'choices' in schema:
        for choice in schema['choices']:
            if isinstance(choice, dict):
                for lang_code in list(choice.keys()):
                    if lang_code not in all_language_codes:
                        del choice[lang_code]
    if schema.get('type') == 'array' and 'items' in schema:
        _parse_schema(schema['items'], path + ['[?]'])
    if schema.get('type') == 'object':
        if isinstance(schema.get('properties'), dict):
            for property_name, property_schema in schema['properties'].items():
                _parse_schema(property_schema, path + [property_name])
                template = _get_dict(property_schema.get('template'))
                if template:
                    template_action = _parse_action_ref(template)
                    property_schema['template'] = template_action


def schema_entry_preprocessor(
        schema: typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any]],
        refs: typing.List[typing.Tuple[str, int]]
) -> None:
    if type(schema) is list:
        for entry in schema:
            schema_entry_preprocessor(entry, refs)
    elif type(schema) is dict:
        if 'type' not in schema.keys():
            for key in schema:
                schema_entry_preprocessor(schema[key], refs)
        if schema.get('type') == 'object':
            template_action_id = schema.get('template')
            if type(template_action_id) is int:
                if ('actions', template_action_id) not in refs:
                    refs.append(('actions', template_action_id))
                action = get_action(template_action_id)
                if action.component is not None and action.fed_id is not None:
                    comp = action.component
                    schema['template'] = ActionRef(
                        action_id=action.fed_id,
                        component_uuid=comp.uuid
                    )
                else:
                    schema['template'] = ActionRef(
                        action_id=action.id,
                        component_uuid=flask.current_app.config['FEDERATION_UUID']
                    )
            schema_properties = schema.get('properties')
            if schema_properties is not None:
                for property in schema_properties:
                    schema_entry_preprocessor(schema['properties'][property], refs)
        if schema.get('type') == 'array':
            schema_entry_preprocessor(schema['items'], refs)



import typing
import dataclasses

from .. import db
from ..models import group_categories, groups, projects
from . import errors


@dataclasses.dataclass(frozen=True)
class GroupCategory:
    """
    This class provides an immutable wrapper around models.group_categories.GroupCategory.
    """
    id: int
    name: typing.Dict[str, str]
    parent_category_id: typing.Optional[int]

    @classmethod
    def from_database(cls, category: group_categories.GroupCategory) -> 'GroupCategory':
        return GroupCategory(
            id=category.id,
            name=category.name,
            parent_category_id=category.parent_category_id
        )


def set_basic_group_categories(
        basic_group_id: int,
        category_ids: typing.Sequence[int]
) -> None:
    """
    Set the categories for a basic group.

    :param basic_group_id: the ID of an existing basic group
    :param category_ids: a list of group category IDs
    :raise errors.GroupDoesNotExistError: is the basic group does not exist
    :raise errors.GroupCategoryDoesNotExistError: if one of the categories does
        not exist
    """
    basic_group = groups.Group.query.filter_by(id=basic_group_id).first()
    if basic_group is None:
        raise errors.GroupDoesNotExistError()
    _set_group_categories(
        group=basic_group,
        category_ids=category_ids
    )


def set_project_group_categories(
        project_group_id: int,
        category_ids: typing.Sequence[int]
) -> None:
    """
    Set the categories for a project group.

    :param project_group_id: the ID of an existing project group
    :param category_ids: a list of group category IDs
    :raise errors.ProjectDoesNotExistError: is the project group does not exist
    :raise errors.GroupCategoryDoesNotExistError: if one of the categories does
        not exist
    """
    project_group = projects.Project.query.filter_by(id=project_group_id).first()
    if project_group is None:
        raise errors.ProjectDoesNotExistError()
    _set_group_categories(
        group=project_group,
        category_ids=category_ids
    )


def _set_group_categories(
        group: typing.Union[projects.Project, groups.Group],
        category_ids: typing.Sequence[int]
) -> None:
    categories = group_categories.GroupCategory.query.filter(group_categories.GroupCategory.id.in_(category_ids)).all()
    if len(categories) != len(set(category_ids)):
        raise errors.GroupCategoryDoesNotExistError()
    group.categories.clear()
    for category in categories:
        group.categories.append(category)
    db.session.add(group)
    db.session.commit()


def create_group_category(
        *,
        name: typing.Dict[str, str],
        parent_category_id: typing.Optional[int] = None
) -> GroupCategory:
    """
    Create a group category.

    :param name: the name translations
    :param parent_category_id: the ID of a parent category, or None
    :raise errors.GroupCategoryDoesNotExistError: if the parent category does
        not exist
    """
    if parent_category_id is not None:
        parent_category = group_categories.GroupCategory.query.filter_by(
            id=parent_category_id
        ).first()
        if parent_category is None:
            raise errors.GroupCategoryDoesNotExistError()
    category = group_categories.GroupCategory(
        name=name,
        parent_category_id=parent_category_id
    )
    db.session.add(category)
    db.session.commit()
    return GroupCategory.from_database(category)


def category_has_ancestor(
        category_id: int,
        potential_ancestor_id: int
) -> bool:
    """
    Check if a category has another category as an ancestor.

    :param category_id: the ID of an existing category
    :param potential_ancestor_id: the ID of an existing category
    :return: whether the potential ancestor category is indeed an ancestor
        of the given category
    :raise errors.GroupCategoryDoesNotExistError: if one of the checked
        categories does not exist
    """
    ancestor_category_id: typing.Optional[int] = category_id
    while ancestor_category_id is not None:
        if ancestor_category_id == potential_ancestor_id:
            return True
        ancestor_category = group_categories.GroupCategory.query.filter_by(id=ancestor_category_id).first()
        if ancestor_category is None:
            raise errors.GroupCategoryDoesNotExistError()
        ancestor_category_id = ancestor_category.parent_category_id
    return False


def update_group_category(
        category_id: int,
        name: typing.Dict[str, str],
        parent_category_id: typing.Optional[int]
) -> None:
    """
    Update a group category.

    :param category_id: the ID of an existing category
    :param name: the name translations
    :param parent_category_id: the ID of a parent category, or None
    :raise errors.GroupCategoryDoesNotExistError: if the category or parent
        category does not exist
    :raise errors.CyclicGroupCategoryError: if the new parent category would
        lead to a cyclic parent-child relationship
    """
    category = group_categories.GroupCategory.query.filter_by(id=category_id).first()
    if category is None:
        raise errors.GroupCategoryDoesNotExistError()
    # ensure this would not create a cyclical parent-child-relationship
    if parent_category_id != category.parent_category_id and parent_category_id is not None and category_has_ancestor(parent_category_id, category_id):
        raise errors.CyclicGroupCategoryError()
    category.name = name
    category.parent_category_id = parent_category_id
    db.session.add(category)
    db.session.commit()


def delete_group_category(
        category_id: int
) -> None:
    """
    Delete a group category.

    :param category_id: the ID of an existing category
    :raise errors.GroupCategoryDoesNotExistError: if the category does not
        exist
    """
    category = group_categories.GroupCategory.query.filter_by(id=category_id).first()
    if category is None:
        raise errors.GroupCategoryDoesNotExistError()
    db.session.delete(category)
    db.session.commit()


def get_group_category(
        category_id: int
) -> GroupCategory:
    """
    Get a group category.

    :param category_id: the ID of an existing category
    :return: the group category
    :raise errors.GroupCategoryDoesNotExistError: if the category does not
        exist
    """
    category = group_categories.GroupCategory.query.filter_by(id=category_id).first()
    if category is None:
        raise errors.GroupCategoryDoesNotExistError()
    return GroupCategory.from_database(category)


def get_group_categories() -> typing.Sequence[GroupCategory]:
    """
    Get all group categories.

    :return: the list of all group categories
    """
    categories = group_categories.GroupCategory.query.order_by(group_categories.GroupCategory.id).all()
    return tuple(
        GroupCategory.from_database(category)
        for category in categories
    )


class GroupCategoryTree(typing.TypedDict):
    basic_group_ids: typing.Set[int]
    project_group_ids: typing.Set[int]
    child_categories: typing.Dict[int, 'GroupCategoryTree']
    contains_basic_groups: bool
    contains_project_groups: bool


def get_group_category_tree(
        basic_group_ids: typing.Optional[typing.Set[int]] = None,
        project_group_ids: typing.Optional[typing.Set[int]] = None
) -> GroupCategoryTree:
    """
    Get the group category tree containing all categories, basic and project groups.

    :return: the group category tree
    """
    categories = get_group_categories()
    basic_groups = groups.Group.query.all()
    project_groups = projects.Project.query.all()
    if basic_group_ids is not None:
        basic_groups = [
            group
            for group in basic_groups
            if group.id in basic_group_ids
        ]
    if project_group_ids is not None:
        project_groups = [
            group
            for group in project_groups
            if group.id in project_group_ids
        ]

    def _build_sub_tree(
            category_id: int,
            categories: typing.Sequence[GroupCategory],
            basic_groups: typing.Sequence[groups.Group],
            project_groups: typing.Sequence[projects.Project]
    ) -> GroupCategoryTree:
        category_tree = GroupCategoryTree(
            basic_group_ids={
                group.id
                for group in basic_groups
                if any(category.id == category_id for category in group.categories)
            },
            project_group_ids={
                group.id
                for group in project_groups
                if any(category.id == category_id for category in group.categories)
            },
            child_categories={
                category.id: _build_sub_tree(category.id, categories, basic_groups, project_groups)
                for category in categories
                if category.parent_category_id == category_id
            },
            contains_basic_groups=False,
            contains_project_groups=False
        )
        category_tree['contains_basic_groups'] = bool(category_tree['basic_group_ids']) or any(
            sub_tree['contains_basic_groups']
            for sub_tree in category_tree['child_categories'].values()
        )
        category_tree['contains_project_groups'] = bool(category_tree['project_group_ids']) or any(
            sub_tree['contains_project_groups']
            for sub_tree in category_tree['child_categories'].values()
        )
        return category_tree

    category_tree = GroupCategoryTree(
        basic_group_ids={
            group.id
            for group in basic_groups
            if not group.categories
        },
        project_group_ids={
            group.id
            for group in project_groups
            if not group.categories
        },
        child_categories={
            category.id: _build_sub_tree(category.id, categories, basic_groups, project_groups)
            for category in categories
            if category.parent_category_id is None
        },
        contains_basic_groups=False,
        contains_project_groups=False
    )
    category_tree['contains_basic_groups'] = bool(category_tree['basic_group_ids']) or any(
        sub_tree['contains_basic_groups']
        for sub_tree in category_tree['child_categories'].values()
    )
    category_tree['contains_project_groups'] = bool(category_tree['project_group_ids']) or any(
        sub_tree['contains_project_groups']
        for sub_tree in category_tree['child_categories'].values()
    )
    return category_tree


def get_full_group_category_names() -> typing.Dict[int, typing.Sequence[typing.Dict[str, str]]]:
    """
    Get the full sequence of ancestor names for all categories.

    :return: a dict containing the names for each category ID
    """
    categories = group_categories.GroupCategory.query.order_by(group_categories.GroupCategory.id).all()
    category_names = {}
    any_missing = True
    while any_missing:
        any_missing = False
        for category in categories:
            if category.id not in category_names:
                if category.parent_category_id is None:
                    category_names[category.id] = [category.name]
                elif category.parent_category_id in category_names:
                    category_names[category.id] = category_names[category.parent_category_id] + [category.name]
                else:
                    any_missing = True
    return {
        category_id: tuple(individual_category_names)
        for category_id, individual_category_names in category_names.items()
    }


def get_full_group_category_name(
        category_id: int
) -> typing.Sequence[typing.Dict[str, str]]:
    """
    Get the full sequence of ancestor names for a category.

    :param category_id: the ID of an existing category
    :return: the names for the category
    :raise errors.GroupCategoryDoesNotExist: if no group category with the given ID exists
    """
    category = get_group_category(category_id=category_id)
    if category.parent_category_id is None:
        return [category.name]
    return list(get_full_group_category_name(category_id=category.parent_category_id)) + [category.name]


def get_basic_group_categories(
        basic_group_id: int
) -> typing.Sequence[GroupCategory]:
    """
    Get the group categories for a basic group.

    :param basic_group_id: the ID of an existing basic group
    :return: the list of group categories
    :raise errors.GroupDoesNotExistError: if then basic group does not exist
    """
    basic_group = groups.Group.query.filter_by(id=basic_group_id).first()
    if basic_group is None:
        raise errors.GroupDoesNotExistError()
    return tuple(
        GroupCategory.from_database(category)
        for category in basic_group.categories
    )


def get_project_group_categories(
        project_group_id: int
) -> typing.Sequence[GroupCategory]:
    """
    Get the group categories for a project group.

    :param project_group_id: the ID of an existing project group
    :return: the list of group categories
    :raise errors.ProjectDoesNotExistError: if then project group does not
        exist
    """
    project_group = projects.Project.query.filter_by(id=project_group_id).first()
    if project_group is None:
        raise errors.ProjectDoesNotExistError()
    return tuple(
        GroupCategory.from_database(category)
        for category in project_group.categories
    )



# coding: utf-8
"""
Add short_description and short_description_is_markdown columns to actions table.
"""

import flask_sqlalchemy

from .utils import table_has_column


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if table_has_column('actions', 'short_description'):
        return False
    if not table_has_column('actions', 'name'):
        return False

    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE actions
        ADD short_description TEXT NOT NULL DEFAULT ''
    """))
    db.session.execute(db.text("""
        ALTER TABLE actions
        ADD short_description_is_markdown BOOLEAN NOT NULL DEFAULT FALSE
    """))
    db.session.execute(db.text("""
        UPDATE actions
        SET short_description = description, short_description_is_markdown=description_is_markdown
    """))
    return True



# coding: utf-8
"""

"""
import sys

import flask_sqlalchemy

from .utils import get_migrations, should_skip_by_index, update_migration_index


def run(db: flask_sqlalchemy.SQLAlchemy) -> None:
    for index, name, function in get_migrations():

        # Skip migration by migration index
        if should_skip_by_index(db, index):
            continue

        try:
            # Perform migration
            if function(db):
                print(f'Migration #{index} "{name}" applied.', file=sys.stderr)
            elif index > 0:
                print(f'Migration #{index} "{name}" skipped.', file=sys.stderr)

            # Update migration index to skip this migration by index in the future
            update_migration_index(db, index)
            db.session.commit()
        except Exception:
            db.session.rollback()
            raise



# coding: utf-8
"""
Add date_format_moment_output column to languages table.
"""

import flask_sqlalchemy

from .utils import table_has_column


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if table_has_column('languages', 'date_format_moment_output'):
        return False

    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE languages
        ADD date_format_moment_output character varying NOT NULL DEFAULT 'll'
    """))
    db.session.execute(db.text("""
        UPDATE languages
        SET date_format_moment_output = 'MMM D, YYYY'
        WHERE id = -99
    """))
    db.session.execute(db.text("""
        UPDATE languages
        SET date_format_moment_output = 'DD.MM.YYYY'
        WHERE id = -98
    """))
    return True



# coding: utf-8
"""
Implementation of generate_placeholder(schema)
"""
import copy
import typing

from flask_login import current_user

from ..errors import UndefinedUnitError, SchemaError, UserDoesNotExistError
from ..schemas.validate import validate
from .utils import get_dimensionality_for_units
from ..users import check_user_exists


def get_default_data(
        schema: typing.Dict[str, typing.Any],
        property_path: typing.Tuple[typing.Union[str, int], ...]
) -> typing.Optional[typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any]]]:
    """
    Return the default data for given schema and property path.

    :param schema: the sampledb object schema
    :param property_path: the path in the schema to generate a value for
    :return: the default data at the given path or None, if there is no default for that path
    """
    current_default_data = generate_placeholder(schema)
    has_default_value = current_default_data is not None
    current_schema: typing.Dict[str, typing.Any] = schema
    for path_element in property_path:
        if current_schema['type'] == 'array':
            current_schema = current_schema.get('items', {})
        else:
            current_schema = current_schema.get('properties', {}).get(path_element)
        if not current_schema:
            return None
        if has_default_value:
            if isinstance(current_default_data, dict) and isinstance(path_element, str):
                current_default_data = current_default_data.get(path_element)
            elif isinstance(current_default_data, list) and isinstance(path_element, int) and 0 <= path_element < len(current_default_data):
                current_default_data = current_default_data[path_element]
            else:
                current_default_data = None
                has_default_value = False
        if not has_default_value:
            current_default_data = generate_placeholder(current_schema)
            has_default_value = current_default_data is not None
    return current_default_data


def generate_placeholder(
        schema: typing.Dict[str, typing.Any],
        path: typing.Optional[typing.List[str]] = None
) -> typing.Optional[typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any]]]:
    """
    Generates a placeholder object based on an object schema.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :return: the generated object or None, if there is no default
    :raise SchemaError: if the type is missing or invalid, or if the schema
        contains a quantity with invalid units
    """
    if path is None:
        path = []
    if 'type' not in schema:
        raise SchemaError('invalid schema', path)
    if schema['type'] == 'array':
        return _generate_array_placeholder(schema, path)
    elif schema['type'] == 'object':
        return _generate_object_placeholder(schema, path)
    elif schema['type'] == 'text':
        return _generate_text_placeholder(schema, path)
    elif schema['type'] == 'datetime':
        return _generate_datetime_placeholder(schema, path)
    elif schema['type'] == 'bool':
        return _generate_bool_placeholder(schema, path)
    elif schema['type'] == 'quantity':
        return _generate_quantity_placeholder(schema, path)
    elif schema['type'] == 'sample':
        return _generate_sample_placeholder(schema, path)  # type: ignore
    elif schema['type'] == 'measurement':
        return _generate_measurement_placeholder(schema, path)  # type: ignore
    elif schema['type'] == 'object_reference':
        return _generate_object_reference_placeholder(schema, path)  # type: ignore
    elif schema['type'] == 'tags':
        return _generate_tags_placeholder(schema, path)
    elif schema['type'] == 'hazards':
        return _generate_hazards_placeholder(schema, path)
    elif schema['type'] == 'user':
        return _generate_user_placeholder(schema, path)
    elif schema['type'] == 'plotly_chart':
        return _generate_plotly_chart_placeholder(schema, path)
    elif schema['type'] == 'timeseries':
        return _generate_timeseries_placeholder(schema, path)
    elif schema['type'] == 'file':
        return _generate_file_placeholder(schema, path)
    else:
        raise SchemaError('invalid type', path)


def _generate_array_placeholder(schema: typing.Dict[str, typing.Any], path: typing.List[str]) -> typing.List[typing.Any]:
    """
    Generates a placeholder array object based on an object schema.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :return: the generated object
    """
    if 'default' in schema:
        return typing.cast(typing.List[typing.Any], schema['default'])
    item_schema = schema['items']
    default_items = schema.get('defaultItems', schema.get('minItems', 0))
    return [
        generate_placeholder(item_schema, path + ['[?]'])
        for _ in range(default_items)
    ]


def _generate_tags_placeholder(schema: typing.Dict[str, typing.Any], path: typing.List[str]) -> typing.Dict[str, typing.Any]:
    """
    Generates a placeholder tags object based on an object schema.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :return: the generated object
    """
    if 'default' in schema:
        return {
            '_type': 'tags',
            'tags': schema['default']
        }
    return {
        '_type': 'tags',
        'tags': []
    }


def _generate_hazards_placeholder(schema: typing.Dict[str, typing.Any], path: typing.List[str]) -> typing.Dict[str, typing.Any]:
    """
    Generate a placeholder GHS hazards object based on an object schema.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :return: the generated object
    """
    return {
        '_type': 'hazards'
    }


def _generate_object_placeholder(schema: typing.Dict[str, typing.Any], path: typing.List[str]) -> typing.Dict[str, typing.Any]:
    """
    Generates a placeholder object based on an object schema.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :return: the generated object
    """
    if 'default' in schema:
        return typing.cast(typing.Dict[str, typing.Any], schema['default'])
    properties = schema['properties']
    return {
        property_name: generate_placeholder(property_schema, path + [property_name])
        for property_name, property_schema in properties.items()
    }


def _generate_bool_placeholder(schema: typing.Dict[str, typing.Any], path: typing.List[str]) -> typing.Union[typing.Dict[str, typing.Any], None]:
    """
    Generates a placeholder boolean object based on an object schema.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :return: the generated object or None, if there is no default value
    """
    if 'default' not in schema:
        return None
    value = schema['default']
    return {
        '_type': 'bool',
        'value': value
    }


def _generate_text_placeholder(schema: typing.Dict[str, typing.Any], path: typing.List[str]) -> typing.Union[typing.Dict[str, typing.Any], None]:
    """
    Generates a placeholder text object based on an object schema.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :return: the generated object or None, if there is no default text
    """
    if 'default' not in schema:
        return None
    text = schema['default']
    return {
        '_type': 'text',
        'text': text
    }


def _generate_datetime_placeholder(schema: typing.Dict[str, typing.Any], path: typing.List[str]) -> typing.Union[typing.Dict[str, typing.Any], None]:
    """
    Generates a placeholder datetime object based on an object schema.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :return: the generated object or None, if there is no default datetime
    """
    if 'default' not in schema or schema['default'] is None:
        return None
    utc_datetime = schema['default']
    return {
        '_type': 'datetime',
        'utc_datetime': utc_datetime
    }


def _generate_quantity_placeholder(schema: typing.Dict[str, typing.Any], path: typing.List[str]) -> typing.Union[typing.Dict[str, typing.Any], None]:
    """
    Generates a placeholder quantity object based on an object schema.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :return: the generated object or None, if there is no default magnitude (in base units)
    :raise SchemaError: if the units are invalid
    """
    if 'default' not in schema:
        return None
    if isinstance(schema['default'], dict):
        default = copy.deepcopy(schema['default'])
        if '_type' not in default:
            default['_type'] = 'quantity'
        validate(default, schema, path)   # fills missing values in quantity in place
        return default
    magnitude_in_base_units = schema['default']
    if isinstance(schema['units'], str):
        units = schema['units']
    else:
        units = schema['units'][0]

    try:
        dimensionality = get_dimensionality_for_units(units)
    except UndefinedUnitError:
        raise SchemaError('invalid units', path)
    return {
        '_type': 'quantity',
        'dimensionality': dimensionality,
        'units': units,
        'magnitude_in_base_units': magnitude_in_base_units
    }


def _generate_sample_placeholder(schema: typing.Dict[str, typing.Any], path: typing.List[str]) -> None:
    """
    Generates a placeholder sample object based on an object schema.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :return: None, as there can be no default sample
    """
    return _generate_object_reference_placeholder(schema, path)


def _generate_measurement_placeholder(schema: typing.Dict[str, typing.Any], path: typing.List[str]) -> None:
    """
    Generates a placeholder measurement object based on an object schema.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :return: None, as there can be no default measurement
    """
    return _generate_object_reference_placeholder(schema, path)


def _generate_object_reference_placeholder(schema: typing.Dict[str, typing.Any], path: typing.List[str]) -> None:
    """
    Generates a placeholder object reference object based on an object schema.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :return: None, as there can be no default object reference
    """
    return None


def _generate_user_placeholder(schema: typing.Dict[str, typing.Any], path: typing.List[str]) -> typing.Optional[typing.Dict[str, typing.Any]]:
    """
    Generates a placeholder user object based on an object schema.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :return: the generated object or None, if there is no default user
    """
    if 'default' in schema:
        if schema['default'] == 'self' and current_user and current_user.is_authenticated:
            user_id = current_user.id
        elif type(schema['default']) is int:
            user_id = schema['default']
        else:
            user_id = None
        if user_id is not None:
            try:
                # ensure the user exists
                check_user_exists(user_id)
            except UserDoesNotExistError:
                pass
            else:
                return {
                    '_type': 'user',
                    'user_id': user_id
                }
    return None


def _generate_plotly_chart_placeholder(schema: typing.Dict[str, typing.Any], path: typing.List[str]) -> typing.Union[typing.Dict[str, typing.Any], None]:
    """
    Generates a placeholder plotly_chart object based on an object schema.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :return: the generated object or None, if there is no default text
    """
    return None


def _generate_timeseries_placeholder(schema: typing.Dict[str, typing.Any], path: typing.List[str]) -> typing.Union[typing.Dict[str, typing.Any], None]:
    """
    Generates a placeholder timeseries object based on an object schema.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :return: None, as there is currently no support for default timeseries
    """
    return None


def _generate_file_placeholder(schema: typing.Dict[str, typing.Any], path: typing.List[str]) -> typing.Optional[typing.Dict[str, typing.Any]]:
    """
    Generates a placeholder file object based on an object schema.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :return: None, as there is currently no support for default files
    """
    return None



# coding: utf-8
"""

"""

import enum
import datetime
import typing

from sqlalchemy.orm import Mapped, Query, relationship

from .files import File
from .. import db
from .utils import Model

if typing.TYPE_CHECKING:
    from .users import User


@enum.unique
class FedUserLogEntryType(enum.Enum):
    OTHER = 0
    IMPORT_USER = 1
    UPDATE_USER = 2
    SHARE_USER = 3
    UPDATE_SHARED_USER = 4
    UPDATE_USER_POLICY = 5
    CREATE_REF_USER = 6


class FedUserLogEntry(Model):
    __tablename__ = 'fed_user_log_entries'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    type: Mapped[FedUserLogEntryType] = db.Column(db.Enum(FedUserLogEntryType), nullable=False)
    user_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=False)
    component_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('components.id'), nullable=False)
    data: Mapped[typing.Dict[str, typing.Any]] = db.Column(db.JSON, nullable=False)
    utc_datetime: Mapped[datetime.datetime] = db.Column(db.TIMESTAMP(timezone=True), nullable=False)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["FedUserLogEntry"]]

    def __init__(
            self,
            type: FedUserLogEntryType,
            user_id: int,
            component_id: int,
            data: typing.Dict[str, typing.Any],
            utc_datetime: typing.Optional[datetime.datetime] = None
    ) -> None:
        super().__init__(
            type=type,
            user_id=user_id,
            component_id=component_id,
            data=data,
            utc_datetime=utc_datetime if utc_datetime is not None else datetime.datetime.now(datetime.timezone.utc)
        )

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id}, type={self.type}, user_id={self.user_id}, utc_datetime={self.utc_datetime}, data={self.data})>'


@enum.unique
class FedObjectLogEntryType(enum.Enum):
    OTHER = 0
    IMPORT_OBJECT = 1
    UPDATE_OBJECT = 2
    ADD_POLICY = 3
    UPDATE_SHARED_OBJECT = 4
    UPDATE_OBJECT_POLICY = 5
    CREATE_REF_OBJECT = 6
    REMOTE_IMPORT_OBJECT = 7


class FedObjectLogEntry(Model):
    __tablename__ = 'fed_object_log_entries'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    type: Mapped[FedObjectLogEntryType] = db.Column(db.Enum(FedObjectLogEntryType), nullable=False)
    object_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('objects_current.object_id'), nullable=False)
    component_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('components.id'), nullable=False)
    data: Mapped[typing.Dict[str, typing.Any]] = db.Column(db.JSON, nullable=False)
    utc_datetime: Mapped[datetime.datetime] = db.Column(db.TIMESTAMP(timezone=True), nullable=False)
    user_id: Mapped[typing.Optional[int]] = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=True)
    user: Mapped[typing.Optional['User']] = relationship('User')

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["FedObjectLogEntry"]]

    def __init__(
            self,
            type: FedObjectLogEntryType,
            object_id: int,
            component_id: int,
            data: typing.Dict[str, typing.Any],
            utc_datetime: typing.Optional[datetime.datetime] = None,
            user_id: typing.Optional[int] = None
    ) -> None:
        super().__init__(
            type=type,
            object_id=object_id,
            component_id=component_id,
            data=data,
            utc_datetime=utc_datetime if utc_datetime is not None else datetime.datetime.now(datetime.timezone.utc),
            user_id=user_id
        )

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id}, type={self.type}, object_id={self.object_id}, utc_datetime={self.utc_datetime}, data={self.data})>'


@enum.unique
class FedLocationLogEntryType(enum.Enum):
    OTHER = 0
    IMPORT_LOCATION = 1
    UPDATE_LOCATION = 2
    SHARE_LOCATION = 3
    UPDATE_SHARED_LOCATION = 4
    UPDATE_LOCATION_POLICY = 5
    CREATE_REF_LOCATION = 6


class FedLocationLogEntry(Model):
    __tablename__ = 'fed_location_log_entries'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    type: Mapped[FedLocationLogEntryType] = db.Column(db.Enum(FedLocationLogEntryType), nullable=False)
    location_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('locations.id'), nullable=False)
    component_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('components.id'), nullable=False)
    data: Mapped[typing.Dict[str, typing.Any]] = db.Column(db.JSON, nullable=False)
    utc_datetime: Mapped[datetime.datetime] = db.Column(db.TIMESTAMP(timezone=True), nullable=False)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["FedLocationLogEntry"]]

    def __init__(
            self,
            type: FedLocationLogEntryType,
            location_id: int,
            component_id: int,
            data: typing.Dict[str, typing.Any],
            utc_datetime: typing.Optional[datetime.datetime] = None
    ) -> None:
        super().__init__(
            type=type,
            location_id=location_id,
            component_id=component_id,
            data=data,
            utc_datetime=utc_datetime if utc_datetime is not None else datetime.datetime.now(datetime.timezone.utc)
        )

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id}, type={self.type}, location_id={self.location_id}, utc_datetime={self.utc_datetime}, data={self.data})>'


@enum.unique
class FedLocationTypeLogEntryType(enum.Enum):
    OTHER = 0
    IMPORT_LOCATION_TYPE = 1
    UPDATE_LOCATION_TYPE = 2
    SHARE_LOCATION_TYPE = 3
    UPDATE_SHARED_LOCATION_TYPE = 4
    UPDATE_LOCATION_TYPE_POLICY = 5
    CREATE_REF_LOCATION_TYPE = 6


class FedLocationTypeLogEntry(Model):
    __tablename__ = 'fed_location_type_log_entries'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    type: Mapped[FedLocationTypeLogEntryType] = db.Column(db.Enum(FedLocationTypeLogEntryType), nullable=False)
    location_type_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('location_types.id'), nullable=False)
    component_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('components.id'), nullable=False)
    data: Mapped[typing.Dict[str, typing.Any]] = db.Column(db.JSON, nullable=False)
    utc_datetime: Mapped[datetime.datetime] = db.Column(db.TIMESTAMP(timezone=True), nullable=False)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["FedLocationTypeLogEntry"]]

    def __init__(
            self,
            type: FedLocationTypeLogEntryType,
            location_type_id: int,
            component_id: int,
            data: typing.Dict[str, typing.Any],
            utc_datetime: typing.Optional[datetime.datetime] = None
    ) -> None:
        super().__init__(
            type=type,
            location_type_id=location_type_id,
            component_id=component_id,
            data=data,
            utc_datetime=utc_datetime if utc_datetime is not None else datetime.datetime.now(datetime.timezone.utc)
        )

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id}, type={self.type}, location_type_id={self.location_type_id}, utc_datetime={self.utc_datetime}, data={self.data})>'


@enum.unique
class FedActionLogEntryType(enum.Enum):
    OTHER = 0
    IMPORT_ACTION = 1
    UPDATE_ACTION = 2
    SHARE_ACTION = 3
    UPDATE_SHARED_ACTION = 4
    UPDATE_ACTION_POLICY = 5
    CREATE_REF_ACTION = 6


class FedActionLogEntry(Model):
    __tablename__ = 'fed_action_log_entries'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    type: Mapped[FedActionLogEntryType] = db.Column(db.Enum(FedActionLogEntryType), nullable=False)
    action_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('actions.id'), nullable=False)
    component_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('components.id'), nullable=False)
    data: Mapped[typing.Dict[str, typing.Any]] = db.Column(db.JSON, nullable=False)
    utc_datetime: Mapped[datetime.datetime] = db.Column(db.TIMESTAMP(timezone=True), nullable=False)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["FedActionLogEntry"]]

    def __init__(
            self,
            type: FedActionLogEntryType,
            action_id: int,
            component_id: int,
            data: typing.Dict[str, typing.Any],
            utc_datetime: typing.Optional[datetime.datetime] = None
    ) -> None:
        super().__init__(
            type=type,
            action_id=action_id,
            component_id=component_id,
            data=data,
            utc_datetime=utc_datetime if utc_datetime is not None else datetime.datetime.now(datetime.timezone.utc)
        )

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id}, type={self.type}, action_id={self.action_id}, utc_datetime={self.utc_datetime}, data={self.data})>'


@enum.unique
class FedActionTypeLogEntryType(enum.Enum):
    OTHER = 0
    IMPORT_ACTION_TYPE = 1
    UPDATE_ACTION_TYPE = 2
    SHARE_ACTION_TYPE = 3
    UPDATE_SHARED_ACTION_TYPE = 4
    UPDATE_ACTION_TYPE_POLICY = 5
    CREATE_REF_ACTION_TYPE = 6


class FedActionTypeLogEntry(Model):
    __tablename__ = 'fed_action_type_log_entries'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    type: Mapped[FedActionTypeLogEntryType] = db.Column(db.Enum(FedActionTypeLogEntryType), nullable=False)
    action_type_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('action_types.id'), nullable=False)
    component_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('components.id'), nullable=False)
    data: Mapped[typing.Dict[str, typing.Any]] = db.Column(db.JSON, nullable=False)
    utc_datetime: Mapped[datetime.datetime] = db.Column(db.TIMESTAMP(timezone=True), nullable=False)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["FedActionTypeLogEntry"]]

    def __init__(
            self,
            type: FedActionTypeLogEntryType,
            action_type_id: int,
            component_id: int,
            data: typing.Dict[str, typing.Any],
            utc_datetime: typing.Optional[datetime.datetime] = None
    ) -> None:
        super().__init__(
            type=type,
            action_type_id=action_type_id,
            component_id=component_id,
            data=data,
            utc_datetime=utc_datetime if utc_datetime is not None else datetime.datetime.now(datetime.timezone.utc)
        )

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id}, type={self.type}, action_type_id={self.action_type_id}, utc_datetime={self.utc_datetime}, data={self.data})>'


@enum.unique
class FedInstrumentLogEntryType(enum.Enum):
    OTHER = 0
    IMPORT_INSTRUMENT = 1
    UPDATE_INSTRUMENT = 2
    SHARE_INSTRUMENT = 3
    UPDATE_SHARED_INSTRUMENT = 4
    UPDATE_INSTRUMENT_POLICY = 5
    CREATE_REF_INSTRUMENT = 6


class FedInstrumentLogEntry(Model):
    __tablename__ = 'fed_instrument_log_entries'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    type: Mapped[FedInstrumentLogEntryType] = db.Column(db.Enum(FedInstrumentLogEntryType), nullable=False)
    instrument_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('instruments.id'), nullable=False)
    component_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('components.id'), nullable=False)
    data: Mapped[typing.Dict[str, typing.Any]] = db.Column(db.JSON, nullable=False)
    utc_datetime: Mapped[datetime.datetime] = db.Column(db.TIMESTAMP(timezone=True), nullable=False)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["FedInstrumentLogEntry"]]

    def __init__(
            self,
            type: FedInstrumentLogEntryType,
            instrument_id: int,
            component_id: int,
            data: typing.Dict[str, typing.Any],
            utc_datetime: typing.Optional[datetime.datetime] = None
    ) -> None:
        super().__init__(
            type=type,
            instrument_id=instrument_id,
            component_id=component_id,
            data=data,
            utc_datetime=utc_datetime if utc_datetime is not None else datetime.datetime.now(datetime.timezone.utc)
        )

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id}, type={self.type}, instrument_id={self.instrument_id}, utc_datetime={self.utc_datetime}, data={self.data})>'


@enum.unique
class FedCommentLogEntryType(enum.Enum):
    OTHER = 0
    IMPORT_COMMENT = 1
    UPDATE_COMMENT = 2
    SHARE_COMMENT = 3
    UPDATE_SHARED_COMMENT = 4
    UPDATE_COMMENT_POLICY = 5
    CREATE_REF_COMMENT = 6


class FedCommentLogEntry(Model):
    __tablename__ = 'fed_comment_log_entries'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    type: Mapped[FedCommentLogEntryType] = db.Column(db.Enum(FedCommentLogEntryType), nullable=False)
    comment_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('comments.id'), nullable=False)
    component_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('components.id'), nullable=False)
    data: Mapped[typing.Dict[str, typing.Any]] = db.Column(db.JSON, nullable=False)
    utc_datetime: Mapped[datetime.datetime] = db.Column(db.TIMESTAMP(timezone=True), nullable=False)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["FedCommentLogEntry"]]

    def __init__(
            self,
            type: FedCommentLogEntryType,
            comment_id: int,
            component_id: int,
            data: typing.Dict[str, typing.Any],
            utc_datetime: typing.Optional[datetime.datetime] = None
    ) -> None:
        super().__init__(
            type=type,
            comment_id=comment_id,
            component_id=component_id,
            data=data,
            utc_datetime=utc_datetime if utc_datetime is not None else datetime.datetime.now(datetime.timezone.utc)
        )

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id}, type={self.type}, comment_id={self.comment_id}, utc_datetime={self.utc_datetime}, data={self.data})>'


@enum.unique
class FedFileLogEntryType(enum.Enum):
    OTHER = 0
    IMPORT_FILE = 1
    UPDATE_FILE = 2
    SHARE_FILE = 3
    UPDATE_SHARED_FILE = 4
    UPDATE_FILE_POLICY = 5
    CREATE_REF_FILE = 6


class FedFileLogEntry(Model):
    __tablename__ = 'fed_file_log_entries'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    type: Mapped[FedFileLogEntryType] = db.Column(db.Enum(FedFileLogEntryType), nullable=False)
    object_id: Mapped[int] = db.Column(db.Integer, nullable=False)
    file_id: Mapped[int] = db.Column(db.Integer, nullable=False)
    component_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('components.id'), nullable=False)
    data: Mapped[typing.Dict[str, typing.Any]] = db.Column(db.JSON, nullable=False)
    utc_datetime: Mapped[datetime.datetime] = db.Column(db.TIMESTAMP(timezone=True), nullable=False)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["FedFileLogEntry"]]

    __table_args__ = (
        db.ForeignKeyConstraint([object_id, file_id], [File.object_id, File.id]),
    )

    def __init__(
            self,
            type: FedFileLogEntryType,
            object_id: int,
            file_id: int,
            component_id: int,
            data: typing.Dict[str, typing.Any],
            utc_datetime: typing.Optional[datetime.datetime] = None
    ) -> None:
        super().__init__(
            type=type,
            object_id=object_id,
            file_id=file_id,
            component_id=component_id,
            data=data,
            utc_datetime=utc_datetime if utc_datetime is not None else datetime.datetime.now(datetime.timezone.utc)
        )

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id}, type={self.type}, file_id={self.file_id}, utc_datetime={self.utc_datetime}, data={self.data})>'


@enum.unique
class FedObjectLocationAssignmentLogEntryType(enum.Enum):
    OTHER = 0
    IMPORT_OBJECT_LOCATION_ASSIGNMENT = 1
    UPDATE_OBJECT_LOCATION_ASSIGNMENT = 2
    SHARE_OBJECT_LOCATION_ASSIGNMENT = 3
    UPDATE_SHARED_OBJECT_LOCATION_ASSIGNMENT = 4
    UPDATE_OBJECT_LOCATION_ASSIGNMENT_POLICY = 5
    CREATE_REF_OBJECT_LOCATION_ASSIGNMENT = 6


class FedObjectLocationAssignmentLogEntry(Model):
    __tablename__ = 'fed_object_location_assignment_log_entries'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    type: Mapped[FedObjectLocationAssignmentLogEntryType] = db.Column(db.Enum(FedObjectLocationAssignmentLogEntryType), nullable=False)
    object_location_assignment_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('object_location_assignments.id'), nullable=False)
    component_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('components.id'), nullable=False)
    data: Mapped[typing.Dict[str, typing.Any]] = db.Column(db.JSON, nullable=False)
    utc_datetime: Mapped[datetime.datetime] = db.Column(db.TIMESTAMP(timezone=True), nullable=False)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["FedObjectLocationAssignmentLogEntry"]]

    def __init__(
            self,
            type: FedObjectLocationAssignmentLogEntryType,
            object_location_assignment_id: int,
            component_id: int,
            data: typing.Dict[str, typing.Any],
            utc_datetime: typing.Optional[datetime.datetime] = None
    ) -> None:
        super().__init__(
            type=type,
            object_location_assignment_id=object_location_assignment_id,
            component_id=component_id,
            data=data,
            utc_datetime=utc_datetime if utc_datetime is not None else datetime.datetime.now(datetime.timezone.utc)
        )

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id}, type={self.type}, object_location_assignment_id={self.object_location_assignment_id}, utc_datetime={self.utc_datetime}, data={self.data})>'



# coding: utf-8
"""

"""
import datetime
import functools
import re
import typing
import sys

import flask
import pytz
from flask_login import current_user

from . import errors
from .. import db
from .background_tasks.send_mail import post_send_mail_task
from .security_tokens import generate_token
from ..models import Authentication, AuthenticationType, User, Tag, BackgroundTaskStatus, BackgroundTask, UserType
from ..utils import ansi_color


def send_user_invitation_email(
        email: str,
        invitation_id: int
) -> BackgroundTaskStatus:
    token_data = {
        'email': email,
        'invitation_id': invitation_id
    }
    token = generate_token(
        token_data,
        salt='invitation',
        secret_key=flask.current_app.config['SECRET_KEY']
    )
    confirm_url = flask.url_for('frontend.invitation_route', token=token, _external=True)

    service_name = flask.current_app.config['SERVICE_NAME']
    subject = service_name + " Invitation"
    html = flask.render_template('mails/user_invitation.html', confirm_url=confirm_url)
    text = flask.render_template('mails/user_invitation.txt', confirm_url=confirm_url)
    while '\n\n\n' in text:
        text = text.replace('\n\n\n', '\n\n')
    return post_send_mail_task(
        subject=subject,
        recipients=[email],
        text=text,
        html=html
    )[0]


def send_email_confirmation_email(
        email: str,
        user_id: int,
        salt: str
) -> typing.Tuple[BackgroundTaskStatus, typing.Optional[BackgroundTask]]:
    token_data = {
        'email': email,
        'user_id': user_id
    }
    token = generate_token(
        token_data,
        salt=salt,
        secret_key=flask.current_app.config['SECRET_KEY']
    )
    confirm_url = flask.url_for("frontend.user_preferences", user_id=user_id, token=token, _external=True)

    service_name = flask.current_app.config['SERVICE_NAME']
    subject = service_name + " Email Confirmation"
    html = flask.render_template('mails/email_confirmation.html', confirm_url=confirm_url)
    text = flask.render_template('mails/email_confirmation.txt', confirm_url=confirm_url)
    while '\n\n\n' in text:
        text = text.replace('\n\n\n', '\n\n')
    return post_send_mail_task(
        subject=subject,
        recipients=[email],
        text=text,
        html=html
    )


def send_recovery_email(
        email: str
) -> typing.Optional[typing.Tuple[BackgroundTaskStatus, typing.Optional[BackgroundTask]]]:
    users = User.query.filter_by(email=email).all()
    email_authentication = Authentication.query.filter(db.and_(Authentication.login['login'].astext == email, Authentication.type == AuthenticationType.EMAIL)).first()
    if email_authentication is not None and email_authentication.user not in users:
        users.append(email_authentication.user)

    # do not provide password resets for users from other databases
    users = [
        user
        for user in users
        if user.type != UserType.FEDERATION_USER
    ]

    if not users:
        return None

    password_reset_urls = {}
    for user in users:
        for authentication_method in user.authentication_methods:
            if authentication_method.type not in {AuthenticationType.LDAP, AuthenticationType.API_TOKEN, AuthenticationType.API_ACCESS_TOKEN}:
                password_reset_urls[authentication_method] = build_confirm_url(authentication_method)

    def filter_printable_authentication_methods(authentication_methods: typing.Iterable[Authentication]) -> typing.List[Authentication]:
        printable_authentication_methods = []
        for authentication_method in authentication_methods:
            if authentication_method.type in {AuthenticationType.LDAP, AuthenticationType.EMAIL, AuthenticationType.OTHER}:
                printable_authentication_methods.append(authentication_method)
        return printable_authentication_methods

    service_name = flask.current_app.config['SERVICE_NAME']
    subject = service_name + " Account Recovery"
    html = flask.render_template('mails/account_recovery.html', email=email, users=users, password_reset_urls=password_reset_urls, filter_printable_authentication_methods=filter_printable_authentication_methods)
    text = flask.render_template('mails/account_recovery.txt', email=email, users=users, password_reset_urls=password_reset_urls, filter_printable_authentication_methods=filter_printable_authentication_methods)
    while '\n\n\n' in text:
        text = text.replace('\n\n\n', '\n\n')
    return post_send_mail_task(
        subject=subject,
        recipients=[email],
        text=text,
        html=html
    )


def send_federated_login_account_creation_email(
    email: str,
    username: str,
    component_id: int,
    fed_id: int
) -> typing.Tuple[BackgroundTaskStatus, typing.Optional[BackgroundTask]]:
    token_data = {
        'email': email,
        'username': username,
        'component_id': component_id,
        'fed_id': fed_id
    }
    token = generate_token(
        token_data,
        salt='federated_login_account',
        secret_key=flask.current_app.config['SECRET_KEY']
    )

    service_name = flask.current_app.config['SERVICE_NAME']
    subject = f"{service_name} Account Creation"
    confirm_url = flask.url_for('.create_account_by_federated_login', token=token, _external=True)

    text = flask.render_template(
        "mails/federated_login_account_creation.txt",
        confirm_url=confirm_url,
        username=token_data['username'],
        service_name=service_name
    )
    html = flask.render_template(
        "mails/federated_login_account_creation.html",
        confirm_url=confirm_url,
        username=token_data['username'],
        service_name=service_name
    )

    return post_send_mail_task(
        subject=subject,
        recipients=[email],
        text=text,
        html=html
    )


def build_confirm_url(
        authentication_method: Authentication,
        salt: str = 'password'
) -> str:
    assert authentication_method.type != AuthenticationType.LDAP

    user_id = authentication_method.user_id
    token = generate_token(authentication_method.id, salt=salt,
                           secret_key=flask.current_app.config['SECRET_KEY'])
    return str(flask.url_for("frontend.user_preferences", user_id=user_id, token=token, _external=True))


def get_translated_text(
        text: typing.Optional[typing.Union[str, typing.Dict[str, str]]],
        language_code: typing.Optional[str] = None,
        default: str = ''
) -> str:
    """
    Return the text in a given language from a translation dictionary.

    If the language does not exist in text, the return value will fall back to
    the english text if it does exist or to an empty string otherwise.

    If text is a string instead of a dict, it will be returned as-is.

    If no language code is provided, the current user's language will be used.

    :param text: a dict mapping language codes to translations
    :param language_code: a language code, or None
    :param default: a text to return if the input text is None or empty
    :return: the translation
    """
    # late import to avoid circular imports
    from .languages import get_user_language
    if isinstance(text, str):
        return text

    if isinstance(text, dict):
        if language_code is None:
            language_code = get_user_language(current_user).lang_code
        translated_text = text.get(language_code)
        if translated_text:
            return translated_text
        # fall back to english
        translated_text = text.get('en')
        if translated_text:
            return translated_text
        # fall back to first language code with non-empty content
        for fallback_language_code in sorted(text):
            translated_text = text[fallback_language_code]
            if translated_text:
                return translated_text

    return default


def get_all_translated_texts(
        text: typing.Optional[typing.Union[str, typing.Dict[str, str]]],
        join_symbol: str,
        default: str = ''
) -> str:
    """
    Return all translations from a translation dictionary.

    The english translation will always be first in the result, if it exists.

    If text is a string instead of a dict, it will be returned as-is.

    If text is neither a string nor a dict, a default will be returned.

    :param text: a dict mapping language codes to translations
    :param join_symbol: the symbol to join translations with
    :param default: a text to return if the input text is None or empty
    :return: the translation
    """
    if isinstance(text, str):
        return text

    if isinstance(text, dict):
        translations = []
        if 'en' in text:
            translations.append(text['en'])
        for language_code, translation in text.items():
            if language_code != 'en':
                translations.append(translation)
        return join_symbol.join(translations)

    return default


def parse_url(
        url: str,
        max_length: int = 2048,
        valid_schemes: typing.Sequence[str] = ('http', 'https', 'ftp', 'file', 'sftp', 'smb')
) -> typing.Dict[str, str]:
    """
    Validate and parse a given URI/URL.

    In case of file-URIs a hostname is required, so local references
    like file:///path and file:/path are considered invalid.

    :param url: string representing the URI to validate
    :param max_length: the URI strings maximum allowed length (default: 2048)
    :param valid_schemes: valid URI schemes
    :return: a dict containing scheme, domain, host, ip_address, port, path and query of the given URI
    :raise errors.InvalidURLError: if the given URL is invalid
    :raise errors.URLTooLongError: if the given URL is too long
    :raise errors.InvalidIPAddressError: if the IP contained in the given URL
        is invalid
    :raise errors.InvalidPortNumberError: if the port contained in the given
        URL is invalid
    """
    if not len(url) <= max_length:
        raise errors.URLTooLongError()

    regex = re.compile(
        # schemes
        r'^(?P<scheme>' + r'|'.join(valid_schemes) + r')://'
        # IP address and port
        r'(?:(?P<ip_address>\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})|'
        # IP address and port
        r'\[(?P<ipv6_address>'
        r'[0-9a-fA-F]{1,4}(?::[0-9a-fA-F]{1,4}){7}|'
        r'(?:[0-9a-fA-F]{1,4}:){1,7}:|'
        r'(?:[0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]|'
        r'(?:[0-9a-fA-F]{1,4}:){1,5}(?::[0-9a-fA-F]{1,4}){1,2}|'
        r'(?:[0-9a-fA-F]{1,4}:){1,4}(?::[0-9a-fA-F]{1,4}){1,3}|'
        r'(?:[0-9a-fA-F]{1,4}:){1,3}(?::[0-9a-fA-F]{1,4}){1,4}|'
        r'(?:[0-9a-fA-F]{1,4}:){1,2}(?::[0-9a-fA-F]{1,4}){1,5}|'
        r'[0-9a-fA-F]{1,4}:(?::[0-9a-fA-F]{1,4}){1,6}|'
        r':(?::[0-9a-fA-F]{1,4}){1,7}'
        r')]|'
        # fqdn
        r'(?P<domain>(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9]{2,}\.?))|'
        # hostname
        r'(?P<host>[A-Z0-9-]+))'
        # port
        r'(?::(?P<port>\d+))?'
        # path
        r'(?P<path>/?|[/?]\S+)'
        # query
        r'(?P<query>\?\S*)?$', re.IGNORECASE)
    match = re.match(regex, url)

    if match is None:
        raise errors.InvalidURLError()

    match_dict = match.groupdict()
    if match_dict['ip_address']:
        for block in match_dict['ip_address'].split('.'):
            num = int(block)
            if num < 0 or num > 225:
                raise errors.InvalidIPAddressError()

    if match_dict['port']:
        num = int(match_dict['port'])
        if num < 1 or num > 65535:
            raise errors.InvalidPortNumberError()

    return match_dict


def print_deprecation_warnings() -> None:
    if show_numeric_tags_warning():
        print(
            ansi_color(
                "Numeric tags are enabled, please evaluate if these are "
                "necessary for your use case and set the configuration value "
                "ENABLE_NUMERIC_TAGS to False to disable them. To learn more,"
                "see: "
                "https://scientific-it-systems.iffgit.fz-juelich.de/SampleDB/administrator_guide/deprecated_features.html#numeric-tags",
                color=33
            ),
            file=sys.stderr,
            end='\n\n'
        )


def show_numeric_tags_warning() -> bool:
    return bool(flask.current_app.config['ENABLE_NUMERIC_TAGS'])


def do_numeric_tags_exist() -> bool:
    """
    Return whether any numeric tags exist in the tags table.

    :return: whether any numeric tags exist
    """
    return Tag.query.filter(Tag.name.op("~")(r'^[0-9\.]+$')).first() is not None


_T = typing.TypeVar('_T')
_CACHE_FUNCTIONS: typing.Set[typing.Callable[..., _T]] = set()


def cache(function: typing.Callable[..., _T]) -> typing.Callable[..., _T]:
    """
    Decorator for adding caching to functions via functools.cache().

    Functions with this decorator can have their cache cleared via the
    clear_cache_functions function.

    :param function: the function to decorate
    :return: the decorated function
    """
    # this decorator will be used before the app is created, so the config
    # variable has to be used directly instead of via flask.current_app.config
    from ..config import ENABLE_FUNCTION_CACHES
    if not ENABLE_FUNCTION_CACHES:
        return function
    cache_function = functools.cache(function)
    _CACHE_FUNCTIONS.add(cache_function)  # type: ignore
    return cache_function


def clear_cache_functions() -> None:
    """
    Clear the cache of all functions decorated with the cache decorator.
    """
    for cache_function in _CACHE_FUNCTIONS:
        cache_function.cache_clear()  # type: ignore


def get_data_and_schema_by_id_path(
        data: typing.Optional[typing.Union[typing.List[typing.Any], typing.Dict[str, typing.Any]]],
        schema: typing.Optional[typing.Dict[str, typing.Any]],
        id_path: typing.List[typing.Union[str, int]],
        convert_id_path_elements: bool = False
) -> typing.Optional[typing.Tuple[typing.Union[typing.List[typing.Any], typing.Dict[str, typing.Any]], typing.Dict[str, typing.Any]]]:
    if data is None or schema is None:
        return None
    if not id_path:
        return data, schema
    if schema.get('type') == 'array':
        if not isinstance(data, list):
            return None
        if not isinstance(id_path[0], int) and convert_id_path_elements:
            try:
                id_path[0] = int(id_path[0])
            except ValueError:
                return None
        if not isinstance(id_path[0], int):
            return None
        if id_path[0] < 0 or id_path[0] >= len(data):
            return None
        return get_data_and_schema_by_id_path(data[id_path[0]], schema['items'], id_path[1:], convert_id_path_elements=convert_id_path_elements)
    if schema.get('type') == 'object':
        if not isinstance(data, dict):
            return None
        if not isinstance(id_path[0], str) and convert_id_path_elements:
            id_path[0] = str(id_path[0])
        if not isinstance(id_path[0], str):
            return None
        if id_path[0] not in schema['properties']:
            return None
        if id_path[0] not in data:
            return None
        return get_data_and_schema_by_id_path(data[id_path[0]], schema['properties'][id_path[0]], id_path[1:], convert_id_path_elements=convert_id_path_elements)
    return None


_application_root_url: typing.Optional[str] = None


def relative_url_for(
        route: str,
        **kwargs: typing.Any
) -> str:
    global _application_root_url
    if _application_root_url is None:
        _application_root_url = flask.url_for('frontend.index')
    kwargs['_external'] = False
    url = flask.url_for(route, **kwargs)
    if url.startswith(_application_root_url):
        url = url[len(_application_root_url):]
    elif url.startswith('/'):
        url = url[1:]
    return url


@functools.cache
def get_postgres_timezone_alias(timezone_name: str, reference_date: datetime.date) -> str:
    current_utc_offset = pytz.timezone(timezone_name).utcoffset(datetime.datetime.now())
    timezone_names_table = db.table(
        'pg_timezone_names',
        db.Column('name', db.String),
        db.Column('utc_offset', db.Interval)
    )
    stmt = db.select(db.exists().where(db.and_(
        timezone_names_table.columns.name == timezone_name,
        timezone_names_table.columns.utc_offset == current_utc_offset
    )))
    if db.session.scalar(stmt):
        return timezone_name
    # create offset-based timezone as a fallback, if timezone data for postgres is out of date or incomplete
    reference_datetime = datetime.datetime(reference_date.year, reference_date.month, reference_date.day)
    naive_reference_datetime = reference_datetime.replace(tzinfo=None)
    reference_utc_offset = pytz.timezone(timezone_name).utcoffset(naive_reference_datetime)
    current_timezone = datetime.timezone(reference_utc_offset)
    current_timezone_name = current_timezone.tzname(naive_reference_datetime)
    # convert from ISO notation to POSIX notation
    if '+' in current_timezone_name:
        current_timezone_name = current_timezone_name.replace('+', '-')
    elif '-' in current_timezone_name:
        current_timezone_name = current_timezone_name.replace('-', '+')
    return current_timezone_name



# coding: utf-8
"""
Logic module for files

Users with WRITE permissions can upload files for samples or measurements. These
files are read only, so to edit an already uploaded file the user will have to
re-upload it. Accordingly, this module only allows creating new files and reading
existing ones.

As multiple files with the same name may exist, a per-object file id is used as
prefix for the actual file names on disk. For each sample or measurement, at
most 10000 files may be uploaded. This limit is arbitrary, but it should be big
enough so that it will not be encountered in practice. As a result, the prefix
will always be 5 characters in length (0000_ to 9999_). The actual file name
may be up to 150 bytes in length (bytes, not characters, encoded as UTF-8).

The files are stored in a directory named after the object's ID, which in turn
will be inside folders named after the action's ID.
"""

import dataclasses
import datetime
import hashlib
import io
import os
import typing

import flask
import requests
from flask_babel import gettext

from . import components, errors, object_log, objects, user_log, users
from .components import get_component
from .errors import FileDoesNotExistError, FileNameTooLongError, \
    InvalidFileStorageError, TooManyFilesForObjectError, FederationFileNotAvailableError
from .objects import get_object
from .users import get_user
from .. import db
from ..models import files
from ..models.file_log import FileLogEntry, FileLogEntryType

MAX_NUM_FILES: int = 10000

SUPPORTED_HASH_ALGORITHMS = ['sha256', 'sha512']
DEFAULT_HASH_ALGORITHM = 'sha256'


@dataclasses.dataclass(frozen=True)
class File:
    """
    This class provides an immutable wrapper around models.files.File.
    """
    id: int
    object_id: int
    user_id: typing.Optional[int]
    data: typing.Dict[str, typing.Any]
    _mutable_file: files.File
    utc_datetime: typing.Optional[datetime.datetime] = None
    fed_id: typing.Optional[int] = None
    component_id: typing.Optional[int] = None
    preview_image_mime_type: typing.Optional[str] = None

    @dataclasses.dataclass(frozen=True)
    class HashInfo:
        algorithm: str
        hexdigest: str

        @classmethod
        def from_binary_data(
                cls,
                algorithm: str,
                binary_data: bytes
        ) -> typing.Optional['File.HashInfo']:
            if algorithm not in SUPPORTED_HASH_ALGORITHMS:
                return None
            return cls(
                algorithm=algorithm,
                hexdigest=getattr(hashlib, algorithm)(binary_data).hexdigest()
            )

    hash: typing.Optional[HashInfo] = None

    @dataclasses.dataclass
    class InfoCache:
        is_hidden: typing.Optional[bool] = None
        hide_reason: typing.Optional[str] = None
        title: typing.Optional[str] = None
        url: typing.Optional[str] = None
        description: typing.Optional[str] = None

    _info_cache: InfoCache = dataclasses.field(default_factory=InfoCache, kw_only=True, repr=False, compare=False)

    @dataclasses.dataclass
    class DataCache:
        binary_data: typing.Optional[bytes] = None
        preview_image_binary_data: typing.Optional[bytes] = None
    _data_cache: DataCache = dataclasses.field(default_factory=DataCache, kw_only=True, repr=False, compare=False)

    @classmethod
    def from_database(cls, file: files.File) -> 'File':
        if file.data is not None:
            data = file.data
        else:
            data = {
                'storage': 'none'
            }
        if isinstance(data.get('hash'), dict):
            hash = File.HashInfo(
                algorithm=data['hash'].get('algorithm'),
                hexdigest=data['hash'].get('hexdigest')
            )
        else:
            hash = None
        return File(
            id=file.id,
            object_id=file.object_id,
            user_id=file.user_id,
            utc_datetime=file.utc_datetime,
            data=data,
            fed_id=file.fed_id,
            component_id=file.component_id,
            hash=hash,
            preview_image_mime_type=file.preview_image_mime_type,
            _mutable_file=file
        )

    @property
    def binary_data(self) -> typing.Optional[bytes]:
        if self._data_cache.binary_data is None:
            self._data_cache.binary_data = self._mutable_file.binary_data
        return self._data_cache.binary_data

    @property
    def preview_image_binary_data(self) -> typing.Optional[bytes]:
        if self._data_cache.preview_image_binary_data is None:
            self._data_cache.preview_image_binary_data = self._mutable_file.preview_image_binary_data
        return self._data_cache.preview_image_binary_data

    @property
    def storage(self) -> str:
        if self.data is None:
            return 'none'
        return str(self.data.get('storage', 'none'))

    @property
    def original_file_name(self) -> str:
        if self.data is not None and self.storage in {'database', 'federation'}:
            return str(self.data.get('original_file_name', ''))
        else:
            raise InvalidFileStorageError()

    @property
    def filepath(self) -> str:
        if self.data is not None and self.storage in {'local_reference'}:
            return str(self.data.get('filepath', ''))
        else:
            raise InvalidFileStorageError()

    @property
    def url(self) -> str:
        if self.data is not None and self.storage == 'url':
            if self._info_cache.url is None:
                log_entry = FileLogEntry.query.filter_by(
                    object_id=self.object_id,
                    file_id=self.id,
                    type=FileLogEntryType.EDIT_URL
                ).order_by(FileLogEntry.utc_datetime.desc()).first()
                if log_entry is not None:
                    self._info_cache.url = log_entry.data['url']
                else:
                    self._info_cache.url = self.data['url']
            return self._info_cache.url
        else:
            raise InvalidFileStorageError()

    @property
    def uploader(self) -> typing.Optional[users.User]:
        if self.user_id is None:
            return None
        return users.get_user(self.user_id)

    @property
    def title(self) -> typing.Optional[str]:
        if self._info_cache.title is None:
            self._info_cache.title = self.real_title
            if self._info_cache.title is None:
                if self.storage in {'database', 'federation'}:
                    return self.original_file_name
                elif self.storage == 'local_reference':
                    return self.filepath
                elif self.storage == 'url':
                    return self.url
                else:
                    raise InvalidFileStorageError()
        return self._info_cache.title

    @property
    def real_title(self) -> typing.Optional[str]:
        log_entry = FileLogEntry.query.filter_by(
            object_id=self.object_id,
            file_id=self.id,
            type=FileLogEntryType.EDIT_TITLE
        ).order_by(FileLogEntry.utc_datetime.desc()).first()
        if log_entry is None:
            return None
        if 'title' not in log_entry.data:
            return None
        return str(log_entry.data['title'])

    @property
    def description(self) -> typing.Union[str, None]:
        if self._info_cache.description is None:
            log_entry = FileLogEntry.query.filter_by(
                object_id=self.object_id,
                file_id=self.id,
                type=FileLogEntryType.EDIT_DESCRIPTION
            ).order_by(FileLogEntry.utc_datetime.desc()).first()
            if log_entry is None:
                return None
            self._info_cache.description = log_entry.data['description']
        return self._info_cache.description

    @property
    def log_entries(self) -> typing.List[FileLogEntry]:
        users_by_id: typing.Dict[typing.Optional[int], typing.Optional[users.User]] = {None: None}
        log_entries = []
        for log_entry in FileLogEntry.query.filter_by(
            object_id=self.object_id,
            file_id=self.id
        ).order_by(FileLogEntry.utc_datetime.asc()).all():
            if log_entry.user_id not in users_by_id:
                users_by_id[log_entry.user_id] = users.get_user(log_entry.user_id)
            log_entry.user = users_by_id[log_entry.user_id]
            log_entries.append(log_entry)
        return log_entries

    @property
    def is_hidden(self) -> bool:
        if self._info_cache.is_hidden is None:
            log_entry = FileLogEntry.query.filter_by(
                object_id=self.object_id,
                file_id=self.id,
                type=FileLogEntryType.HIDE_FILE
            ).order_by(FileLogEntry.utc_datetime.desc()).first()
            if log_entry is None:
                self._info_cache.is_hidden = False
                return False
            hide_time = log_entry.utc_datetime
            self._info_cache.hide_reason = log_entry.data['reason']
            log_entry = FileLogEntry.query.filter_by(
                object_id=self.object_id,
                file_id=self.id,
                type=FileLogEntryType.UNHIDE_FILE
            ).order_by(FileLogEntry.utc_datetime.desc()).first()
            if log_entry is None:
                self._info_cache.is_hidden = True
                return True
            unhide_time = log_entry.utc_datetime
            self._info_cache.is_hidden = hide_time > unhide_time
        return bool(self._info_cache.is_hidden)

    @property
    def hide_reason(self) -> typing.Optional[str]:
        if self.is_hidden:
            return self._info_cache.hide_reason
        return None

    def open(self, read_only: bool = True) -> typing.BinaryIO:
        if self.storage == 'database':
            if self.binary_data is not None:
                return io.BytesIO(self.binary_data)
            else:
                return io.BytesIO(b'')
        elif self.storage == 'federation':
            object = get_object(self.object_id)
            if self.component_id:
                component = get_component(self.component_id)
            else:
                raise InvalidFileStorageError()
            from .federation.update import get_binary
            try:
                file_data = get_binary(f'/federation/v1/shares/objects/{object.fed_object_id}/files/{self.fed_id}', component)
            except errors.UnauthorizedRequestError:
                raise FederationFileNotAvailableError()
            except errors.MissingComponentAddressError:
                raise FederationFileNotAvailableError()
            except errors.RequestServerError:
                raise FederationFileNotAvailableError()
            except errors.RequestError:
                raise FederationFileNotAvailableError()
            except requests.exceptions.ConnectionError:
                raise FederationFileNotAvailableError()
            except errors.NoAuthenticationMethodError:
                raise FederationFileNotAvailableError()
            return file_data
        else:
            raise InvalidFileStorageError()


def create_local_file_reference(
        object_id: int,
        user_id: int,
        filepath: str,
        hash: typing.Optional[File.HashInfo] = None
) -> File:
    """
    Create a file as a link to a local file and add it to the object and user
    logs.

    :param object_id: the ID of an existing object
    :param user_id: the ID of an existing user
    :param filepath: the filepath relative to the mounting directory of the
        download service
    :param hash: the hash info for this file
    :return: the newly created file
    :raise errors.ObjectDoesNotExistError: when no object with the given
        object ID exists
    :raise errors.UserDoesNotExistError: when no user with the given user ID
        exists
    :raise errors.TooManyFilesForObjectError: when there are already 10000
        files for the object with the given id
    :raise errors.UnauthorizedRequestError: when the current user is not allowed to
        add this certain filepath
    """
    path_permissions: typing.Dict[str, typing.List[int]] = flask.current_app.config['DOWNLOAD_SERVICE_WHITELIST']
    user = get_user(user_id)

    filepath = os.path.normpath(filepath)
    for path, user_ids in path_permissions.items():
        if not path.endswith(os.sep):
            path += os.sep
        if filepath.startswith(path):
            if user.is_admin or user_id in user_ids:
                break
    else:
        raise errors.UnauthorizedRequestError()

    db_file = _create_db_file(
        object_id=object_id,
        user_id=user_id,
        data={
            'storage': 'local_reference',
            'filepath': filepath,
            'valid': True,  # to distinguish from local file references created before the whitelist was checked properly
            'hash': {
                'algorithm': hash.algorithm,
                'hexdigest': hash.hexdigest
            } if hash is not None else None
        }
    )
    file = File.from_database(db_file)
    _create_file_logs(file)
    return file


def create_url_file(
        object_id: int,
        user_id: int,
        url: str,
        *,
        create_log_entry: bool = True,
        utc_datetime: typing.Optional[datetime.datetime] = None
) -> File:
    """
    Create a file as a link to a URL and add it to the object and user logs.

    :param object_id: the ID of an existing object
    :param user_id: the ID of an existing user
    :param url: the file URL
    :param create_log_entry: whether to create a log entry
    :param utc_datetime: the datetime for the file, or None
    :return: the newly created file
    :raise errors.ObjectDoesNotExistError: when no object with the given
        object ID exists
    :raise errors.UserDoesNotExistError: when no user with the given user ID
        exists
    :raise errors.TooManyFilesForObjectError: when there are already 10000
        files for the object with the given id
    """
    db_file = _create_db_file(
        object_id=object_id,
        user_id=user_id,
        data={
            'storage': 'url',
            'url': url
        },
        utc_datetime=utc_datetime
    )
    file = File.from_database(db_file)
    if create_log_entry:
        _create_file_logs(file)
    return file


@typing.overload
def create_database_file(
        object_id: int,
        user_id: int,
        file_name: str,
        save_content: typing.Callable[[typing.BinaryIO], None],
        hash: typing.Optional[File.HashInfo] = None,
        *,
        create_log_entry: bool = True,
        preview_image_binary_data: bytes,
        preview_image_mime_type: str,
        utc_datetime: typing.Optional[datetime.datetime] = None
) -> File:
    ...


@typing.overload
def create_database_file(
        object_id: int,
        user_id: int,
        file_name: str,
        save_content: typing.Callable[[typing.BinaryIO], None],
        hash: typing.Optional[File.HashInfo] = None,
        *,
        create_log_entry: bool = True,
        preview_image_binary_data: None = None,
        preview_image_mime_type: None = None,
        utc_datetime: typing.Optional[datetime.datetime] = None
) -> File:
    ...


def create_database_file(
        object_id: int,
        user_id: int,
        file_name: str,
        save_content: typing.Callable[[typing.BinaryIO], None],
        hash: typing.Optional[File.HashInfo] = None,
        *,
        create_log_entry: bool = True,
        preview_image_binary_data: typing.Optional[bytes] = None,
        preview_image_mime_type: typing.Optional[str] = None,
        utc_datetime: typing.Optional[datetime.datetime] = None
) -> File:
    """
    Create a new database file and add it to the object and user logs.

    The function will call save_content with an opened file (in binary mode).

    :param object_id: the ID of an existing object
    :param user_id: the ID of an existing user
    :param file_name: the original file name
    :param save_content: a function which will save the file's content to the
        given stream. The function will be called at most once.
    :param hash: the hash info for this file
    :param create_log_entry: whether to create a log entry
    :param preview_image_binary_data: binary data for the preview image, or None
    :param preview_image_mime_type: mime type for the preview image, or None
    :param utc_datetime: the datetime for the file, or None
    :return: the newly created file
    :raise errors.ObjectDoesNotExistError: when no object with the given
        object ID exists
    :raise errors.UserDoesNotExistError: when no user with the given user ID
        exists
    :raise errors.FileNameTooLongError: when the file name is longer than 150
        bytes when encoded as UTF-8
    """
    # ensure that the file name is valid
    if len(file_name.encode('utf8')) > 150:
        raise FileNameTooLongError()

    binary_data_file = io.BytesIO()
    save_content(binary_data_file)
    binary_data = binary_data_file.getvalue()
    if hash is None:
        hash = File.HashInfo.from_binary_data(
            algorithm=DEFAULT_HASH_ALGORITHM,
            binary_data=binary_data
        )

    db_file = _create_db_file(
        object_id=object_id,
        user_id=user_id,
        data={
            'storage': 'database',
            'original_file_name': file_name,
            'hash': {
                'algorithm': hash.algorithm,
                'hexdigest': hash.hexdigest
            } if hash is not None else None
        },
        utc_datetime=utc_datetime
    )
    db_file.binary_data = binary_data
    db_file.preview_image_binary_data = preview_image_binary_data
    db_file.preview_image_mime_type = preview_image_mime_type.lower() if preview_image_mime_type else ''
    db.session.commit()
    file = File.from_database(db_file)
    if create_log_entry:
        _create_file_logs(file)
    return file


def get_mutable_file(
        file_id: int,
        object_id: int,
        component_id: typing.Optional[int] = None
) -> files.File:
    """
    :param file_id: the federated ID of the file
    :param object_id: ID of the object the requested file is assigned to
    :param component_id: the components ID (source)
    :return: the file
    """
    db_file: typing.Optional[files.File]
    if component_id is None:
        db_file = files.File.query.filter_by(id=file_id, object_id=object_id).first()
    else:
        db_file = files.File.query.filter_by(fed_id=file_id, object_id=object_id, component_id=component_id).first()
    if db_file is None:
        objects.check_object_exists(object_id)
        if component_id is not None:
            components.check_component_exists(component_id)
        raise errors.FileDoesNotExistError
    return db_file


def get_file(
        file_id: int,
        object_id: int,
        component_id: typing.Optional[int] = None
) -> File:
    """
    :param file_id: the federated ID of the file
    :param object_id: ID of the object the requested file is assigned to
    :param component_id: the components ID (source)
    :return: the file
    """
    return File.from_database(get_mutable_file(file_id=file_id, object_id=object_id, component_id=component_id))


def create_fed_file(
        object_id: int,
        user_id: typing.Optional[int],
        data: typing.Optional[typing.Dict[str, typing.Any]],
        save_content: typing.Optional[typing.Callable[[typing.BinaryIO], None]],
        utc_datetime: datetime.datetime,
        fed_id: int,
        component_id: int
) -> files.File:
    """
    Creates a new file referencing a federated file.

    :param object_id: the ID of an existing object
    :param user_id: the ID of an existing user
    :param data: data describing the file or file reference
    :param save_content: a function which will save the file's content to the
        given stream. The function will be called at most once.
    :param utc_datetime: the time of file-creation
    :param fed_id: the federated ID of the file
    :param component_id: the components ID (source)
    :return: the created File object
    :raise errors.ComponentDoesNotExistError: when no component with the given
        component ID exists
    """
    file = _create_db_file(object_id, user_id, data, utc_datetime, fed_id, component_id)
    if save_content:
        binary_data_file = io.BytesIO()
        save_content(binary_data_file)
        binary_data = binary_data_file.getvalue()
        file.binary_data = binary_data
        db.session.commit()
    return file


def _create_file_logs(file: File) -> None:
    """
    Create object and user logs for a new file.

    :param file: the file to create log entries for
    """
    file_id = file.id
    object_id = file.object_id
    user_id = file.user_id
    if user_id is not None:
        object_log.upload_file(user_id=user_id, object_id=object_id, file_id=file_id)
        user_log.upload_file(user_id=user_id, object_id=object_id, file_id=file_id)


def _create_db_file(
        object_id: int,
        user_id: typing.Optional[int],
        data: typing.Optional[typing.Dict[str, typing.Any]],
        utc_datetime: typing.Optional[datetime.datetime] = None,
        fed_id: typing.Optional[int] = None,
        component_id: typing.Optional[int] = None
) -> files.File:
    """
    Creates a new file in the database.

    :param object_id: the ID of an existing object
    :param user_id: the ID of an existing user
    :param data: the file storage data
    :param utc_datetime: the time of file-creation, only if file
        has been imported
    :param fed_id: the federated ID of the file, only if file
        has been imported
    :param component_id: the components ID (source), only if file
        has been imported
    :raise errors.ObjectDoesNotExistError: when no object with the given
        object ID exists
    :raise errors.UserDoesNotExistError: when no user with the given user ID
        exists
    :raise errors.TooManyFilesForObjectError: when there are already 10000
        files for the object with the given id
    :raise errors.ComponentDoesNotExistError: when there is no component with the
        given ID
    """
    if (component_id is not None and (fed_id is None or utc_datetime is None)) or (component_id is None and (user_id is None or data is None or fed_id is not None)):
        raise TypeError('Invalid parameter combination.')

    # ensure that the object exists
    object = objects.get_object(object_id)
    if user_id is not None:
        # ensure that the user exists
        users.check_user_exists(user_id)
    if component_id is not None:
        # ensure that the component exists
        components.check_component_exists(component_id)
    # calculate the next file id
    previous_file_id = db.session.query(db.func.max(files.File.id)).filter(files.File.object_id == object.id).scalar()
    if previous_file_id is None:
        file_id = 0
    else:
        file_id = previous_file_id + 1
    if file_id >= MAX_NUM_FILES:
        raise TooManyFilesForObjectError()
    db_file = files.File(
        file_id=file_id,
        object_id=object_id,
        user_id=user_id,
        data=data,
        utc_datetime=utc_datetime,
        fed_id=fed_id,
        component_id=component_id
    )
    db.session.add(db_file)
    db.session.commit()
    return db_file


def update_file_information(object_id: int, file_id: int, user_id: int, title: str, description: str, url: typing.Optional[str] = None) -> None:
    """
    Creates new file log entries for updating a file's information.

    :param object_id: the ID of an existing object
    :param file_id: the ID of a file for the object
    :param user_id: the ID of an existing user
    :param title: the new title
    :param description: the new description
    :param url: the new url
    :raise errors.FileDoesNotExistError: when no file with the given object ID
        and file ID exists
    """
    file = get_file_for_object(object_id=object_id, file_id=file_id)
    if file is None:
        raise FileDoesNotExistError()
    if not title:
        if file.storage == 'url':
            title = file.url
        else:
            title = file.original_file_name
    if title != file.real_title:
        log_entry = FileLogEntry(type=FileLogEntryType.EDIT_TITLE, object_id=object_id, file_id=file_id, user_id=user_id, data={
            'title': title
        })
        db.session.add(log_entry)
        db.session.commit()
    if url and file.storage == 'url' and url != file.url:
        log_entry = FileLogEntry(type=FileLogEntryType.EDIT_URL, object_id=object_id, file_id=file_id, user_id=user_id, data={
            'url': url
        })
        db.session.add(log_entry)
        db.session.commit()
    if description != file.description and not (description == "" and file.description is None):
        log_entry = FileLogEntry(type=FileLogEntryType.EDIT_DESCRIPTION, object_id=object_id, file_id=file_id, user_id=user_id, data={
            'description': description
        })
        db.session.add(log_entry)
        db.session.commit()


def hide_file(
        object_id: int,
        file_id: int,
        user_id: int,
        reason: str,
        utc_datetime: typing.Optional[datetime.datetime] = None
) -> None:
    """
    Hides a file.

    :param object_id: the ID of an existing object
    :param file_id: the ID of a file for the object
    :param user_id: the ID of an existing user
    :param reason: the reason for hiding the file
    :param utc_datetime: the time when the file was hidden or None to select the current time
    :raise errors.FileDoesNotExistError: when no file with the given object ID
        and file ID exists
    """
    file = get_file_for_object(object_id=object_id, file_id=file_id)
    if file is None:
        raise FileDoesNotExistError()
    if not reason:
        reason = ''
    log_entry = FileLogEntry(type=FileLogEntryType.HIDE_FILE, object_id=object_id, file_id=file_id, user_id=user_id, data={
        'reason': reason
    }, utc_datetime=utc_datetime)
    db.session.add(log_entry)
    db.session.commit()


def get_file_for_object(object_id: int, file_id: int) -> typing.Optional[File]:
    """
    Returns the file with the given file ID for the object with the given object ID.

    :param object_id: the ID of an existing object
    :param file_id: the ID of an existing file for the object
    :return: the file or None
    """
    db_file = files.File.query.filter_by(object_id=object_id, id=file_id).first()
    if db_file is None:
        return None
    return File.from_database(db_file)


def get_files_for_object(object_id: int) -> typing.List[File]:
    """
    Returns a list of files for an object.

    :param object_id: the ID of an existing object
    :return: the list of files, sorted by upload time from first to last
    :raise errors.ObjectDoesNotExistError: when no object with the given
        object ID exists
    """
    db_files = files.File.query.filter_by(object_id=object_id).order_by(db.asc(files.File.utc_datetime)).all()
    if not db_files:
        # ensure that the object exists
        objects.check_object_exists(object_id)
    return [File.from_database(db_file) for db_file in db_files]


def get_file_names_by_id_for_object(
        object_id: int
) -> typing.Dict[int, typing.Tuple[str, str]]:
    files = get_files_for_object(object_id)
    file_names_by_id = {}
    for file in files:
        if file.storage in {'database', 'federation'}:
            actual_name = file.original_file_name
            if file.is_hidden:
                display_name = gettext('Hidden File') + f' #{file.id}'
            else:
                display_name = actual_name
            file_names_by_id[file.id] = (actual_name, display_name)
    return file_names_by_id



# coding: utf-8
"""

"""
import typing

import flask
import flask_login
import itsdangerous
from flask_babel import _
from flask_wtf import FlaskForm
from wtforms import BooleanField, PasswordField, SelectField, SelectMultipleField
from wtforms.validators import InputRequired

from . import frontend
from .. import logic
from ..utils import object_permissions_required, FlaskResponseT
from ..models import Permissions, DataverseExportStatus, BackgroundTaskStatus

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


class DataverseExportForm(FlaskForm):
    dataverse = SelectField(
        choices=[],
        validators=[InputRequired()]
    )
    files = SelectMultipleField()
    tags = SelectMultipleField()
    api_token = PasswordField()


class DataverseAPITokenForm(FlaskForm):
    api_token = PasswordField(validators=[InputRequired()])
    store_api_token = BooleanField()


@frontend.route('/objects/<int:object_id>/dataverse_export/', methods=['GET', 'POST'])
@object_permissions_required(Permissions.GRANT)
def dataverse_export(object_id: int) -> FlaskResponseT:
    dataverse_export_status = logic.dataverse_export.get_dataverse_export_state(object_id)
    existing_dataverse_url = logic.dataverse_export.get_dataverse_url(object_id)

    if dataverse_export_status:
        if dataverse_export_status == DataverseExportStatus.TASK_CREATED:
            flask.flash(_('The object is currently being exported.'), 'info')
            return flask.redirect(flask.url_for('.object', object_id=object_id))
        elif dataverse_export_status == DataverseExportStatus.EXPORT_FINISHED and existing_dataverse_url:
            return flask.render_template(
                'objects/dataverse_export_already_exported.html',
                object_id=object_id,
                dataverse_url=existing_dataverse_url
            )

    server_url = flask.current_app.config['DATAVERSE_URL']
    if not server_url:
        return flask.abort(404)

    object = logic.objects.get_object(object_id)
    if object.eln_import_id is not None:
        return flask.abort(403)
    if object.component_id is not None:
        flask.flash(_('Exporting imported objects is not supported.'), 'error')
        return flask.redirect(flask.url_for('.object', object_id=object_id))

    dataverse_export_form = DataverseExportForm()

    user_id = flask_login.current_user.id
    api_token = logic.dataverse_export.get_user_valid_api_token(server_url, user_id)
    had_invalid_api_token = False
    dataverse_api_token_form = DataverseAPITokenForm()
    if not api_token:
        if dataverse_api_token_form.validate_on_submit():
            form_api_token = dataverse_api_token_form.api_token.data.strip()
            api_token_valid = logic.dataverse_export.is_api_token_valid(server_url, form_api_token)
            if api_token_valid:
                api_token = form_api_token
                if dataverse_api_token_form.store_api_token.data:
                    logic.settings.set_user_settings(user_id, {'DATAVERSE_API_TOKEN': api_token})
                else:
                    dataverse_export_form.api_token.data = api_token
            else:
                had_invalid_api_token = True
                api_token = None
        elif dataverse_export_form.validate_on_submit():
            form_api_token = dataverse_export_form.api_token.data.strip()
            api_token_valid = logic.dataverse_export.is_api_token_valid(server_url, form_api_token)
            if api_token_valid:
                api_token = form_api_token
            else:
                had_invalid_api_token = True
                api_token = None
    if not api_token:
        return flask.render_template(
            'objects/dataverse_export_api_token.html',
            object_id=object_id,
            dataverse_api_token_form=dataverse_api_token_form,
            had_invalid_api_token=had_invalid_api_token
        )

    if object.data:
        properties = {
            tuple(whitelist_path): metadata
            for metadata, whitelist_path, title_path in logic.dataverse_export.flatten_metadata(object.data)
            if not ('_type' in metadata and metadata['_type'] == 'tags')
        }
    else:
        properties = {}

    root_dataverses = flask.current_app.config['DATAVERSE_ROOT_IDS'].split(',')
    candidate_dataverses = []
    for dataverse in root_dataverses:
        dataverse_info = logic.dataverse_export.get_dataverse_info(server_url, api_token, dataverse)
        if dataverse_info is not None:
            candidate_dataverses.append((0, dataverse_info))
            candidate_dataverses += logic.dataverse_export.list_dataverses(server_url, api_token, dataverse)

    dataverses = []
    for level, dataverse in candidate_dataverses:
        dataverses.append((
            level,
            dataverse,
            logic.dataverse_export.dataverse_has_required_metadata_blocks(server_url, api_token, dataverse['id'])
        ))

    dataverse_export_form.dataverse.choices = [
        (str(dataverse['id']), '\u00A0' * (2 * level) + dataverse['title'])
        for level, dataverse, enabled in dataverses
        if enabled
    ]
    if not dataverse_export_form.dataverse.choices:
        flask.flash(_('No suitable Dataverses available for export.'), 'error')

    dataverse_export_form.files.choices = [
        (str(file.id), file.original_file_name)
        for file in logic.files.get_files_for_object(object_id)
        if file.storage == 'database'
    ]

    if object.data:
        tags_set = set()
        for property in object.data.values():
            if '_type' in property and property['_type'] == 'tags':
                for tag in property['tags']:
                    tags_set.add(tag)
        tags = list(tags_set)
        tags.sort()
    else:
        tags = []

    dataverse_export_form.tags.choices = [
        (tag, tag)
        for tag in tags
    ]

    export_properties: typing.List[typing.List[typing.Union[int, str]]] = [
        ['name']
    ]
    for whitelist_path in properties:
        if 'property,' + ','.join(map(str, whitelist_path)) in flask.request.form:
            export_properties.append(list(whitelist_path))
    if dataverse_export_form.validate_on_submit():
        file_id_whitelist_set = set()
        for file_id in dataverse_export_form.files.data:
            try:
                file_id = int(file_id)
            except ValueError:
                continue
            file_id_whitelist_set.add(file_id)
        file_id_whitelist = list(file_id_whitelist_set)
        file_id_whitelist.sort()
        tag_whitelist_set = set()
        for tag in dataverse_export_form.tags.data:
            if tag in tags:
                tag_whitelist_set.add(tag)
        tag_whitelist = list(tag_whitelist_set)
        tag_whitelist.sort()
        dataverse = dataverse_export_form.dataverse.data
        task_status_or_result, task = logic.background_tasks.background_dataverse_export.post_dataverse_export_task(
            object_id=object_id,
            user_id=flask_login.current_user.id,
            server_url=server_url,
            api_token=api_token,
            dataverse=dataverse,
            property_whitelist=export_properties,
            file_id_whitelist=file_id_whitelist,
            tag_whitelist=tag_whitelist
        )
        if task:
            serializer = itsdangerous.URLSafeSerializer(secret_key=flask.current_app.config['SECRET_KEY'], salt='dataverse-export-task')
            token = serializer.dumps((flask_login.current_user.id, task.id))
            return flask.redirect(flask.url_for(".dataverse_export_loading", task_id=task.id, token=token))
        else:
            success, export_result = typing.cast(typing.Tuple[bool, dict[str, typing.Any]], task_status_or_result)
            if success:
                return flask.redirect(export_result['url'])
            else:
                error_message = export_result['error_message']
                flask.flash(_('An error occurred while exporting the object: %(error)s', error=_(error_message, service_name=flask.current_app.config["SERVICE_NAME"], dataverse_name=flask.current_app.config["DATAVERSE_NAME"])), 'error')
                return flask.redirect(flask.url_for(".dataverse_export", object_id=object_id))

    return flask.render_template(
        "objects/dataverse_export.html",
        properties=properties,
        export_properties=export_properties,
        dataverses=dataverses,
        object_id=object_id,
        get_title_for_property=lambda path: logic.dataverse_export.get_title_for_property(path, object.schema) if object.schema else '?',
        get_property_export_default=lambda path: logic.dataverse_export.get_property_export_default(path, object.schema) if object.schema else False,
        dataverse_export_form=dataverse_export_form
    )


@frontend.route("/objects/<int:task_id>/dataverse_export_loading/", methods=['GET'])
@flask_login.login_required
def dataverse_export_loading(task_id: int) -> FlaskResponseT:
    serializer = itsdangerous.URLSafeSerializer(secret_key=flask.current_app.config['SECRET_KEY'], salt='dataverse-export-task')
    token = flask.request.args.get('token', '')
    try:
        if serializer.loads(token) != [flask_login.current_user.id, task_id]:
            return flask.abort(403)
    except itsdangerous.BadData:
        return flask.abort(403)
    task_result = logic.background_tasks.get_background_task_result(task_id, False)
    if task_result:
        if task_result["status"].is_final():
            if task_result["status"] == BackgroundTaskStatus.DONE:
                return flask.redirect(task_result["result"])
            else:
                return flask.render_template('objects/dataverse_export_loading.html', polling=False, error_message=task_result['result'])
        else:
            return flask.render_template('objects/dataverse_export_loading.html', task_id=task_id, polling=True, token=token), 202
    else:
        return flask.abort(404)


@frontend.route("/objects/<int:task_id>/dataverse_export_status/", methods=['GET'])
@flask_login.login_required
def dataverse_export_status(task_id: int) -> FlaskResponseT:
    serializer = itsdangerous.URLSafeSerializer(secret_key=flask.current_app.config['SECRET_KEY'], salt='dataverse-export-task')
    token = flask.request.args.get('token', '')
    try:
        if serializer.loads(token) != [flask_login.current_user.id, task_id]:
            return flask.abort(403)
    except itsdangerous.BadData:
        return flask.abort(403)
    task_result = logic.background_tasks.get_background_task_result(task_id, False)
    if task_result:
        if task_result["status"].is_final():
            if task_result["status"] == BackgroundTaskStatus.DONE:
                return flask.jsonify({
                    "dataverse_url": task_result["result"]
                })
            else:
                return flask.jsonify({
                    "url": flask.url_for(".dataverse_export_loading", task_id=task_id, token=token),
                    "error_message": task_result["result"]
                }), 406

        else:
            return flask.jsonify({}), 202
    else:
        return flask.jsonify({
            "url": flask.url_for(".dataverse_export_loading", task_id=task_id, token=token),
            "error_message": "task_id not found"
        }), 404
